# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import abc
import copy

from azure.cli.core import azclierror
from azure.cli.core.commands.arm import add_usage, remove_usage, set_usage
from knack.arguments import CLICommandArgument, CaseInsensitiveList
from knack.preview import PreviewItem
from knack.experimental import ExperimentalItem
from knack.util import status_tag_messages
from knack.log import get_logger

from ._arg_action import AAZSimpleTypeArgAction, AAZObjectArgAction, AAZDictArgAction, \
    AAZListArgAction, AAZGenericUpdateAction, AAZGenericUpdateForceStringAction, AAZAnyTypeArgAction
from ._base import AAZBaseType, AAZUndefined
from ._field_type import AAZObjectType, AAZStrType, AAZIntType, AAZBoolType, AAZFloatType, AAZListType, AAZDictType, \
    AAZSimpleType, AAZFreeFormDictType, AAZAnyType
from ._field_value import AAZObject
from ._arg_fmt import AAZObjectArgFormat, AAZListArgFormat, AAZDictArgFormat, AAZFreeFormDictArgFormat, \
    AAZSubscriptionIdArgFormat, AAZResourceLocationArgFormat, AAZResourceIdArgFormat, AAZUuidFormat, AAZDateFormat, \
    AAZTimeFormat, AAZDateTimeFormat, AAZDurationFormat, AAZFileArgTextFormat, AAZPaginationTokenArgFormat, \
    AAZIntArgFormat
from .exceptions import AAZUnregisteredArg
from ._prompt import AAZPromptInput

# pylint: disable=redefined-builtin, protected-access, too-few-public-methods, too-many-instance-attributes

logger = get_logger(__name__)


class AAZArgumentsSchema(AAZObjectType):
    """ Arguments' schema should be defined as fields of it """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._fmt = AAZObjectArgFormat()

    def __call__(self, data=None):
        return AAZObject(
            schema=self,
            data=self.process_data(data=data)
        )


class AAZArgEnum:
    """Argument enum properties"""

    def __init__(self, items, case_sensitive=False, support_extension=False):
        self._case_sensitive = case_sensitive
        self.support_extension = support_extension
        self.items = items

    def to_choices(self):
        """Generate choices property of CLICommandArgument"""
        choices = list(self.items)
        if not self._case_sensitive:
            choices = CaseInsensitiveList(choices)
        return choices

    def __getitem__(self, data):
        key = data
        if isinstance(self.items, dict):
            # convert data, it can be key, value or key.lower() when not case sensitive
            for k, v in self.items.items():
                if v == data or k == data or not self._case_sensitive and k.lower() == data.lower():
                    key = k
                    break

        if key in self.items:
            if isinstance(self.items, (list, tuple, set)):
                return key
            if isinstance(self.items, dict):
                return self.items[key]
            raise NotImplementedError()
        if self.support_extension:
            # support extension value which is not in choices
            if isinstance(self.items, dict):
                values = list(self.items.values())
            elif isinstance(self.items, (list, tuple, set)):
                values = list(self.items)
            try:
                data_type = type(values[0])
                value = data_type(data)
                logger.warning("Use extended value '%s' outside choices %s.", str(value), self.to_choices())
                return value
            except (ValueError, IndexError):
                pass
        raise azclierror.InvalidArgumentValueError(
            f"unrecognized value '{data}' from choices '{self.to_choices()}' ")


class AAZBaseArg(AAZBaseType):
    """Base argument"""

    def __init__(self, options=None, required=False, help=None, arg_group=None, is_preview=False, is_experimental=False,
                 id_part=None, default=AAZUndefined, blank=AAZUndefined, nullable=False, fmt=None, registered=True,
                 configured_default=None, completer=None):
        """

        :param options: argument optional names.
        :param required: argument is required or not.
        :param help: help can be either string (for short-summary) or a dict with `name`, `short-summary`,
                    `long-summary` and `populator-commands` properties
        :param arg_group: argument group.
        :param is_preview: is preview.
        :param is_experimental: is experimental.
        :param id_part: whether needs id part support in azure.cli.core
        :param default: when the argument flag is not appeared, the default value will be used.
        :param blank: when the argument flag is used without value data, the blank value will be used.
        :param nullable: argument can accept `None` as value
        :param fmt: argument format
        :param registered: control whether register argument into command display
        :param configured_default: the key to retrieve the default value from cli configuration
        :param completer: tab completion if completion is active
        """
        super().__init__(options=options, nullable=nullable)
        self._help = {}  # the key in self._help can be 'name', 'short-summary', 'long-summary', 'populator-commands'

        if self._options:
            self._help['name'] = ' '.join(self._options)
            if isinstance(help, str):
                self._help['short-summary'] = help
            elif isinstance(help, dict):
                self._help.update(help)

        self._required = required
        self._arg_group = arg_group
        self._is_preview = is_preview
        self._is_experimental = is_experimental
        self._id_part = id_part
        self._default = default
        self._blank = blank
        self._fmt = fmt
        self._registered = registered
        self._configured_default = configured_default
        self._completer = completer

    def to_cmd_arg(self, name, **kwargs):
        """ convert AAZArg to CLICommandArgument """
        if not self._registered:
            # argument will not registered in command display
            raise AAZUnregisteredArg()

        options_list = [*self._options] if self._options else None
        arg = CLICommandArgument(
            dest=name,
            options_list=options_list,
            # if default is not None, arg is not required.
            required=self._required if self._default == AAZUndefined else False,
            help=self._help.get('short-summary', None),
            id_part=self._id_part,
            default=copy.deepcopy(self._default),
        )
        if self._arg_group:
            arg.arg_group = self._arg_group

        if self._blank != AAZUndefined:
            arg.nargs = '?'
            if isinstance(self._blank, AAZPromptInput):
                short_summary = arg.type.settings.get('help', None) or ''
                if short_summary:
                    short_summary += '  '
                short_summary += self._blank.help_message
                arg.help = short_summary

        cli_ctx = kwargs.get('cli_ctx', None)
        if cli_ctx is None:
            # define mock cli_ctx for PreviewItem and ExperimentalItem
            class _CLI_CTX:
                enable_color = False
            cli_ctx = _CLI_CTX

        if options_list is None:
            target = f"--{name.replace('_', '-')}"
        else:
            target = sorted(options_list, key=len)[-1]

        if self._is_preview:
            def _get_preview_arg_message(self):
                subject = "{} '{}'".format(self.object_type.capitalize(), self.target)
                return status_tag_messages['preview'].format(subject)
            arg.preview_info = PreviewItem(
                cli_ctx=cli_ctx,
                target=target,
                object_type="argument",
                message_func=_get_preview_arg_message
            )

        if self._is_experimental:
            def _get_experimental_arg_message(self):
                # "Argument xxx"
                subject = "{} '{}'".format(self.object_type.capitalize(), self.target)
                return status_tag_messages['experimental'].format(subject)

            arg.experimental_info = ExperimentalItem(
                cli_ctx=cli_ctx,
                target=target,
                object_type="argument",
                message_func=_get_experimental_arg_message
            )

        if self._configured_default:
            arg.configured_default = self._configured_default

        if self._completer:
            from azure.cli.core.decorators import Completer
            assert isinstance(self._completer, Completer)
            arg.completer = self._completer

        action = self._build_cmd_action()   # call sub class's implementation to build CLICommandArgument action
        if action:
            arg.action = action

        return arg

    @abc.abstractmethod
    def _build_cmd_action(self):
        """build argparse Action"""
        raise NotImplementedError()

    @property
    def _type_in_help(self):
        """argument type displayed in help"""
        return "Undefined"


class AAZSimpleTypeArg(AAZBaseArg, AAZSimpleType):
    """Argument accept simple value"""

    def __init__(self, enum=None, enum_case_sensitive=False, enum_support_extension=False, **kwargs):
        super().__init__(**kwargs)
        self.enum = AAZArgEnum(
            enum, case_sensitive=enum_case_sensitive, support_extension=enum_support_extension
        ) if enum else None

    def to_cmd_arg(self, name, **kwargs):
        arg = super().to_cmd_arg(name, **kwargs)
        if self.enum:
            choices = self.enum.to_choices()
            if self.enum.support_extension:
                # Display the allowed values in help only without verifying the user input in argparse
                short_summary = arg.type.settings.get('help', None) or ''
                if short_summary:
                    short_summary += '  '
                short_summary += 'Allowed values: {}.'.format(', '.join(sorted([str(x) for x in choices])))
                arg.help = short_summary
            else:
                # this will verify the user input in argparse
                arg.choices = choices   # convert it's enum value into choices in arg
        return arg

    def _build_cmd_action(self):
        class Action(AAZSimpleTypeArgAction):
            _schema = self  # bind action class with current schema

        return Action


class AAZStrArg(AAZSimpleTypeArg, AAZStrType):

    @property
    def _type_in_help(self):
        return "String"


class AAZDurationArg(AAZStrArg):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZDurationFormat()
        super().__init__(fmt=fmt, **kwargs)

    @property
    def _type_in_help(self):
        return "Duration"


class AAZDateArg(AAZStrArg):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZDateFormat()
        super().__init__(fmt=fmt, **kwargs)

    @property
    def _type_in_help(self):
        return "Date"


class AAZTimeArg(AAZStrArg):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZTimeFormat()
        super().__init__(fmt=fmt, **kwargs)

    @property
    def _type_in_help(self):
        return "Time"


class AAZDateTimeArg(AAZStrArg):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZDateTimeFormat()
        super().__init__(fmt=fmt, **kwargs)

    @property
    def _type_in_help(self):
        return "DateTime"


class AAZUuidArg(AAZStrArg):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZUuidFormat()
        super().__init__(fmt=fmt, **kwargs)

    @property
    def _type_in_help(self):
        return "GUID/UUID"


class AAZPasswordArg(AAZStrArg):

    @property
    def _type_in_help(self):
        return "Password"


class AAZIntArg(AAZSimpleTypeArg, AAZIntType):

    @property
    def _type_in_help(self):
        return "Int"


class AAZBoolArg(AAZSimpleTypeArg, AAZBoolType):

    def __init__(self, blank=True, enum=None, **kwargs):
        enum = enum or {
            'true': True, 't': True, 'yes': True, 'y': True, '1': True,
            "false": False, 'f': False, 'no': False, 'n': False, '0': False,
        }
        super().__init__(blank=blank, enum=enum, **kwargs)

    @property
    def _type_in_help(self):
        return "Boolean"


class AAZFloatArg(AAZSimpleTypeArg, AAZFloatType):

    @property
    def _type_in_help(self):
        return "Float"


class AAZAnyTypeArg(AAZBaseArg, AAZAnyType):

    def _build_cmd_action(self):
        class Action(AAZAnyTypeArgAction):
            _schema = self  # bind action class with current schema
        return Action

    def to_cmd_arg(self, name, **kwargs):
        from ._help import shorthand_help_messages
        arg = super().to_cmd_arg(name, **kwargs)
        short_summary = arg.type.settings.get('help', None) or ''
        if short_summary:
            short_summary += '  '
        short_summary += shorthand_help_messages['short-summary-anytype']
        arg.help = short_summary
        return arg

    @property
    def _type_in_help(self):
        return "Any"


class AAZCompoundTypeArg(AAZBaseArg):

    @abc.abstractmethod
    def _build_cmd_action(self):
        raise NotImplementedError()

    def to_cmd_arg(self, name, **kwargs):
        from ._help import shorthand_help_messages
        arg = super().to_cmd_arg(name, **kwargs)
        short_summary = arg.type.settings.get('help', None) or ''
        if short_summary:
            short_summary += '  '
        short_summary += shorthand_help_messages['short-summary'] + ' ' + shorthand_help_messages['show-help']
        if isinstance(self, AAZListArg) and self.singular_options:
            singular_options = '  Singular flags: ' + ' '.join(
                [f'`{opt}`' for opt in self.singular_options])
            short_summary += singular_options + '.'
        arg.help = short_summary
        return arg


class AAZObjectArg(AAZCompoundTypeArg, AAZObjectType):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZObjectArgFormat()
        super().__init__(fmt=fmt, **kwargs)

    def to_cmd_arg(self, name, **kwargs):
        arg = super().to_cmd_arg(name, **kwargs)
        if self._blank != AAZUndefined:
            arg.nargs = '*'
        else:
            arg.nargs = '+'
        return arg

    def _build_cmd_action(self):
        class Action(AAZObjectArgAction):
            _schema = self  # bind action class with current schema

        return Action

    @property
    def _type_in_help(self):
        return "Object"


class AAZDictArg(AAZCompoundTypeArg, AAZDictType):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZDictArgFormat()
        super().__init__(fmt=fmt, **kwargs)

    def to_cmd_arg(self, name, **kwargs):
        arg = super().to_cmd_arg(name, **kwargs)
        if self._blank != AAZUndefined:
            arg.nargs = '*'
        else:
            arg.nargs = '+'
        return arg

    def _build_cmd_action(self):
        class Action(AAZDictArgAction):
            _schema = self  # bind action class with current schema

        return Action

    @property
    def _type_in_help(self):
        return f"Dict<String,{self.Element._type_in_help}>"


# Warning: This type should not be used any more, the new aaz-dev-tools only use AAZDictType with AAZAnyType
class AAZFreeFormDictArg(AAZDictArg, AAZFreeFormDictType):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZFreeFormDictArgFormat()
        super().__init__(fmt=fmt, **kwargs)
        # for backward compatible, support nullable value here for AAZFreeFormDictArg,
        # from the new code gen tools, it will avoid using AAZFreeFormDictArg
        self._element = AAZAnyTypeArg(nullable=True)


class AAZListArg(AAZCompoundTypeArg, AAZListType):

    def __init__(self, fmt=None, singular_options=None, **kwargs):
        fmt = fmt or AAZListArgFormat()
        super().__init__(fmt=fmt, **kwargs)
        self.singular_options = singular_options

    def to_cmd_arg(self, name, **kwargs):
        arg = super().to_cmd_arg(name, **kwargs)
        if self.singular_options:
            assert arg.options_list
            arg.options_list.extend(self.singular_options)  # support to parse singular options

        if self._blank != AAZUndefined:
            arg.nargs = '*'
        else:
            arg.nargs = '+'
        return arg

    def _build_cmd_action(self):
        class Action(AAZListArgAction):
            _schema = self  # bind action class with current schema

        return Action

    @property
    def _type_in_help(self):
        return f"List<{self.Element._type_in_help}>"


class AAZResourceGroupNameArg(AAZStrArg):

    def __init__(
            self, options=('--resource-group', '-g'), id_part='resource_group',
            help="Name of resource group. "
                 "You can configure the default group using `az configure --defaults group=<name>`",
            configured_default='group',
            completer=None,
            **kwargs):
        from azure.cli.core.commands.parameters import get_resource_group_completion_list
        completer = completer or get_resource_group_completion_list
        super().__init__(
            options=options,
            id_part=id_part,
            help=help,
            configured_default=configured_default,
            completer=completer,
            **kwargs
        )

    def to_cmd_arg(self, name, **kwargs):
        from azure.cli.core.local_context import LocalContextAttribute, LocalContextAction, ALL
        arg = super().to_cmd_arg(name, **kwargs)
        arg.local_context_attribute = LocalContextAttribute(
            name='resource_group_name',
            actions=[LocalContextAction.SET, LocalContextAction.GET],
            scopes=[ALL]
        )
        return arg


class AAZResourceLocationArg(AAZStrArg):

    def __init__(
            self, options=('--location', '-l'),
            help="Location. Values from: `az account list-locations`. "
                 "You can configure the default location using `az configure --defaults location=<location>`.",
            fmt=None,
            configured_default='location',
            completer=None,
            **kwargs):
        from azure.cli.core.commands.parameters import get_location_completion_list

        fmt = fmt or AAZResourceLocationArgFormat()
        completer = completer or get_location_completion_list
        super().__init__(
            options=options,
            help=help,
            fmt=fmt,
            configured_default=configured_default,
            completer=completer,
            **kwargs
        )

    def to_cmd_arg(self, name, **kwargs):
        from azure.cli.core.local_context import LocalContextAttribute, LocalContextAction, ALL
        arg = super().to_cmd_arg(name, **kwargs)
        if self._required and \
                isinstance(self._fmt, AAZResourceLocationArgFormat) and self._fmt._resource_group_arg is not None:
            # when location is required and it will be retrived from resource group by default, arg is not required.
            arg.required = False
            short_summary = arg.type.settings.get('help', None) or ''
            if short_summary:
                short_summary += '  '
            short_summary += "When not specified, the location of the resource group will be used."
            arg.help = short_summary

        arg.local_context_attribute = LocalContextAttribute(
            name='location',
            actions=[LocalContextAction.SET, LocalContextAction.GET],
            scopes=[ALL]
        )
        return arg


class AAZResourceIdArg(AAZStrArg):
    """ResourceId Argument"""

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZResourceIdArgFormat()
        super().__init__(fmt=fmt, **kwargs)


class AAZSubscriptionIdArg(AAZStrArg):

    def __init__(
            self, help="Name or ID of subscription. You can configure the default subscription "
                       "using `az account set -s NAME_OR_ID`",
            fmt=None,
            completer=None,
            **kwargs):
        from azure.cli.core._completers import get_subscription_id_list
        fmt = fmt or AAZSubscriptionIdArgFormat()
        completer = completer or get_subscription_id_list
        super().__init__(
            help=help,
            fmt=fmt,
            completer=completer,
            **kwargs
        )


class AAZFileArg(AAZStrArg):

    def __init__(self, fmt=None, **kwargs):
        fmt = fmt or AAZFileArgTextFormat()
        super().__init__(fmt=fmt, **kwargs)


# Generic Update arguments
class AAZGenericUpdateForceStringArg(AAZBoolArg):

    def __init__(
            self, options=('--force-string',), arg_group='Generic Update',
            help="When using 'set' or 'add', preserve string literals instead of attempting to convert to JSON.",
            **kwargs):
        super().__init__(
            options=options,
            help=help,
            arg_group=arg_group,
            **kwargs,
        )

    def _build_cmd_action(self):
        class Action(AAZGenericUpdateForceStringAction):
            _schema = self  # bind action class with current schema
        return Action


class AAZGenericUpdateArg(AAZBaseArg, AAZListType):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.Element = AAZStrType()

    @abc.abstractmethod
    def _build_cmd_action(self):
        raise NotImplementedError()


class AAZGenericUpdateSetArg(AAZGenericUpdateArg):

    def __init__(
            self, options=('--set',), arg_group='Generic Update',
            help='Update an object by specifying a property path and value to set.'
                 '  Example: {}'.format(set_usage),
            **kwargs):
        super().__init__(
            options=options,
            help=help,
            arg_group=arg_group,
            **kwargs,
        )

    def to_cmd_arg(self, name, **kwargs):
        arg = super().to_cmd_arg(name, **kwargs)
        arg.nargs = '+'
        arg.metavar = 'KEY=VALUE'
        return arg

    def _build_cmd_action(self):
        class Action(AAZGenericUpdateAction):
            ACTION_NAME = "set"
        return Action


class AAZGenericUpdateAddArg(AAZGenericUpdateArg):

    def __init__(
            self, options=('--add',), arg_group='Generic Update',
            help='Add an object to a list of objects by specifying a path and key value pairs.'
                 '  Example: {}'.format(add_usage),
            **kwargs):
        super().__init__(
            options=options,
            help=help,
            arg_group=arg_group,
            **kwargs,
        )

    def to_cmd_arg(self, name, **kwargs):
        arg = super().to_cmd_arg(name, **kwargs)
        arg.nargs = '+'
        arg.metavar = 'LIST KEY=VALUE'
        return arg

    def _build_cmd_action(self):
        class Action(AAZGenericUpdateAction):
            ACTION_NAME = "add"
        return Action


class AAZGenericUpdateRemoveArg(AAZGenericUpdateArg):

    def __init__(
            self, options=('--remove', ), arg_group='Generic Update',
            help='Remove a property or an element from a list.'
                 '  Example: {}'.format(remove_usage),
            **kwargs):
        super().__init__(
            options=options,
            help=help,
            arg_group=arg_group,
            **kwargs,
        )

    def to_cmd_arg(self, name, **kwargs):
        arg = super().to_cmd_arg(name, **kwargs)
        arg.nargs = '+'
        arg.metavar = 'LIST INDEX'
        return arg

    def _build_cmd_action(self):
        class Action(AAZGenericUpdateAction):
            ACTION_NAME = "remove"
        return Action


class AAZPaginationTokenArg(AAZStrArg):
    def __init__(
            self, options=("--next-token",), arg_group="Pagination",
            help="Token to specify where to start paginating. This is the token value from a previously truncated "
                 "response.",
            fmt=None,
            **kwargs
    ):
        fmt = fmt or AAZPaginationTokenArgFormat()

        super().__init__(
            options=options,
            arg_group=arg_group,
            help=help,
            fmt=fmt,
            **kwargs,
        )


class AAZPaginationLimitArg(AAZIntArg):
    def __init__(
            self, options=("--max-items",), arg_group="Pagination",
            help="Total number of items to return in the command's output. If the total number of items available is "
                 "more than the value specified, a token is provided in the command's output. To resume pagination, "
                 "provide the token value in `--next-token` argument of a subsequent command.",
            fmt=None,
            **kwargs
    ):
        fmt = fmt or AAZIntArgFormat(minimum=1)

        super().__init__(
            options=options,
            arg_group=arg_group,
            help=help,
            fmt=fmt,
            **kwargs,
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "compute-recommender spot-placement-score"
)
class SpotPlacementScore(AAZCommand):
    """Generate placement scores for Spot VM skus.

    :example: generate spot vm placement score example
        az compute-recommender spot-placement-score -l eastus --subscription ffffffff-ffff-ffff-ffff-ffffffffffff --availability-zones true --desired-locations '["eastus", "eastus2"]' --desired-count 1 --desired-sizes '[{"sku": "Standard_D2_v2"}]'
    """

    _aaz_info = {
        "version": "2025-06-05",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/providers/microsoft.compute/locations/{}/placementscores/spot/generate", "2025-06-05"],
        ]
    }

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.location = AAZResourceLocationArg(
            required=True,
            id_part="name",
        )

        # define Arg Group "SpotPlacementScoresInput"

        _args_schema = cls._args_schema
        _args_schema.availability_zones = AAZBoolArg(
            options=["--availability-zones"],
            arg_group="SpotPlacementScoresInput",
            help="Defines if the scope is zonal or regional.",
        )
        _args_schema.desired_count = AAZIntArg(
            options=["--desired-count"],
            arg_group="SpotPlacementScoresInput",
            help="Desired instance count per region/zone based on the scope.",
        )
        _args_schema.desired_locations = AAZListArg(
            options=["--desired-locations"],
            arg_group="SpotPlacementScoresInput",
            help="The desired regions",
            required=True,
        )
        _args_schema.desired_sizes = AAZListArg(
            options=["--desired-sizes"],
            arg_group="SpotPlacementScoresInput",
            help="The desired resource SKUs.",
            required=True,
        )

        desired_locations = cls._args_schema.desired_locations
        desired_locations.Element = AAZStrArg()

        desired_sizes = cls._args_schema.desired_sizes
        desired_sizes.Element = AAZObjectArg()

        _element = cls._args_schema.desired_sizes.Element
        _element.sku = AAZStrArg(
            options=["sku"],
            help="The resource's CRP virtual machine SKU size.",
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        self.SpotPlacementScoresPost(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class SpotPlacementScoresPost(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/providers/Microsoft.Compute/locations/{location}/placementScores/spot/generate",
                **self.url_parameters
            )

        @property
        def method(self):
            return "POST"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "location", self.ctx.args.location,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2025-06-05",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("availabilityZones", AAZBoolType, ".availability_zones")
            _builder.set_prop("desiredCount", AAZIntType, ".desired_count")
            _builder.set_prop("desiredLocations", AAZListType, ".desired_locations", typ_kwargs={"flags": {"required": True}})
            _builder.set_prop("desiredSizes", AAZListType, ".desired_sizes", typ_kwargs={"flags": {"required": True}})

            desired_locations = _builder.get(".desiredLocations")
            if desired_locations is not None:
                desired_locations.set_elements(AAZStrType, ".")

            desired_sizes = _builder.get(".desiredSizes")
            if desired_sizes is not None:
                desired_sizes.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".desiredSizes[]")
            if _elements is not None:
                _elements.set_prop("sku", AAZStrType, ".sku")

            return self.serialize_content(_content_value)

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.availability_zones = AAZBoolType(
                serialized_name="availabilityZones",
            )
            _schema_on_200.desired_count = AAZIntType(
                serialized_name="desiredCount",
            )
            _schema_on_200.desired_locations = AAZListType(
                serialized_name="desiredLocations",
            )
            _schema_on_200.desired_sizes = AAZListType(
                serialized_name="desiredSizes",
            )
            _schema_on_200.placement_scores = AAZListType(
                serialized_name="placementScores",
            )

            desired_locations = cls._schema_on_200.desired_locations
            desired_locations.Element = AAZStrType()

            desired_sizes = cls._schema_on_200.desired_sizes
            desired_sizes.Element = AAZObjectType()

            _element = cls._schema_on_200.desired_sizes.Element
            _element.sku = AAZStrType()

            placement_scores = cls._schema_on_200.placement_scores
            placement_scores.Element = AAZObjectType()

            _element = cls._schema_on_200.placement_scores.Element
            _element.availability_zone = AAZStrType(
                serialized_name="availabilityZone",
            )
            _element.is_quota_available = AAZBoolType(
                serialized_name="isQuotaAvailable",
            )
            _element.region = AAZStrType()
            _element.score = AAZStrType()
            _element.sku = AAZStrType()

            return cls._schema_on_200


class _SpotPlacementScoreHelper:
    """Helper class for SpotPlacementScore"""


__all__ = ["SpotPlacementScore"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long,too-many-lines

from azure.cli.core.commands import DeploymentOutputLongRunningOperation
from azure.cli.core.commands.arm import (
    deployment_validate_table_format, handle_template_based_exception)
from azure.cli.core.commands import CliCommandType

from azure.cli.command_modules.network._format import (
    transform_local_gateway_table_output, transform_dns_record_set_output,
    transform_dns_zone_table_output, transform_public_ip_create_output,
    transform_traffic_manager_create_output,
    transform_geographic_hierachy_table_output,
    transform_service_community_table_output, transform_waf_rule_sets_table_output,
    transform_network_usage_table, transform_nsg_rule_table_output,
    transform_vnet_table_output, transform_effective_route_table, transform_effective_nsg,
    transform_vnet_gateway_routes_table, transform_vnet_gateway_bgp_peer_table)
from azure.cli.command_modules.network._validators import (
    process_ag_create_namespace,
    process_lb_create_namespace,
    process_nw_flow_log_show_namespace,
    process_public_ip_create_namespace,
    process_vpn_connection_create_namespace,
    process_appgw_waf_policy_update, process_cross_region_lb_create_namespace)

NETWORK_VROUTER_DEPRECATION_INFO = 'network routeserver'
NETWORK_VROUTER_PEERING_DEPRECATION_INFO = 'network routeserver peering'


# pylint: disable=too-many-locals, too-many-statements
def load_command_table(self, _):
    # region NetworkRoot
    with self.command_group('network'):
        from azure.cli.command_modules.network.custom import UsagesList
        self.command_table['network list-usages'] = UsagesList(loader=self,
                                                               table_transformer=transform_network_usage_table)
    # endregion

    # region ApplicationGateways
    with self.command_group("network application-gateway") as g:
        from .custom import ApplicationGatewayUpdate
        self.command_table["network application-gateway update"] = ApplicationGatewayUpdate(loader=self)
        g.custom_command("show-backend-health", "show_ag_backend_health")
        g.custom_command("create", "create_application_gateway",
                         transform=DeploymentOutputLongRunningOperation(self.cli_ctx),
                         supports_no_wait=True,
                         table_transformer=deployment_validate_table_format,
                         validator=process_ag_create_namespace,
                         exception_handler=handle_template_based_exception)

    with self.command_group("network application-gateway address-pool"):
        from .custom import AddressPoolCreate, AddressPoolUpdate
        self.command_table["network application-gateway address-pool create"] = AddressPoolCreate(loader=self)
        self.command_table["network application-gateway address-pool update"] = AddressPoolUpdate(loader=self)

    with self.command_group("network application-gateway auth-cert"):
        from .custom import AuthCertCreate, AuthCertUpdate
        self.command_table["network application-gateway auth-cert create"] = AuthCertCreate(loader=self)
        self.command_table["network application-gateway auth-cert update"] = AuthCertUpdate(loader=self)

    with self.command_group("network application-gateway root-cert"):
        from .custom import RootCertCreate, RootCertUpdate
        self.command_table["network application-gateway root-cert create"] = RootCertCreate(loader=self)
        self.command_table["network application-gateway root-cert update"] = RootCertUpdate(loader=self)

    with self.command_group("network application-gateway client-cert"):
        from .custom import ClientCertAdd, ClientCertUpdate
        self.command_table["network application-gateway client-cert add"] = ClientCertAdd(loader=self)
        self.command_table["network application-gateway client-cert update"] = ClientCertUpdate(loader=self)

    with self.command_group("network application-gateway frontend-ip"):
        from .custom import FrontedIPCreate, FrontedIPUpdate
        self.command_table["network application-gateway frontend-ip create"] = FrontedIPCreate(loader=self)
        self.command_table["network application-gateway frontend-ip update"] = FrontedIPUpdate(loader=self)

    with self.command_group("network application-gateway settings"):
        from .custom import SettingsCreate, SettingsUpdate
        self.command_table["network application-gateway settings create"] = SettingsCreate(loader=self)
        self.command_table["network application-gateway settings update"] = SettingsUpdate(loader=self)

    with self.command_group("network application-gateway http-settings"):
        from .custom import HTTPSettingsCreate, HTTPSettingsUpdate
        self.command_table["network application-gateway http-settings create"] = HTTPSettingsCreate(loader=self)
        self.command_table["network application-gateway http-settings update"] = HTTPSettingsUpdate(loader=self)

    with self.command_group("network application-gateway listener"):
        from .custom import ListenerCreate, ListenerUpdate
        self.command_table["network application-gateway listener create"] = ListenerCreate(loader=self)
        self.command_table["network application-gateway listener update"] = ListenerUpdate(loader=self)

    with self.command_group("network application-gateway http-listener"):
        from .custom import HTTPListenerCreate, HTTPListenerUpdate
        self.command_table["network application-gateway http-listener create"] = HTTPListenerCreate(loader=self)
        self.command_table["network application-gateway http-listener update"] = HTTPListenerUpdate(loader=self)

    with self.command_group("network application-gateway identity") as g:
        from .custom import IdentityAssign
        self.command_table["network application-gateway identity assign"] = IdentityAssign(loader=self)
        g.custom_command("remove", "remove_ag_identity", supports_no_wait=True)

    with self.command_group("network application-gateway private-link"):
        from .custom import AGPrivateLinkAdd, AGPrivateLinkRemove, AGPrivateLinkIPConfigAdd
        self.command_table["network application-gateway private-link add"] = AGPrivateLinkAdd(loader=self)
        self.command_table["network application-gateway private-link remove"] = AGPrivateLinkRemove(loader=self)
        self.command_table["network application-gateway private-link ip-config add"] = AGPrivateLinkIPConfigAdd(loader=self)

    with self.command_group("network application-gateway redirect-config"):
        from .custom import RedirectConfigCreate, RedirectConfigUpdate
        self.command_table["network application-gateway redirect-config create"] = RedirectConfigCreate(loader=self)
        self.command_table["network application-gateway redirect-config update"] = RedirectConfigUpdate(loader=self)

    with self.command_group("network application-gateway rewrite-rule"):
        from .custom import AGRewriteRuleCreate, AGRewriteRuleUpdate
        self.command_table["network application-gateway rewrite-rule create"] = AGRewriteRuleCreate(loader=self)
        self.command_table["network application-gateway rewrite-rule update"] = AGRewriteRuleUpdate(loader=self)

    with self.command_group("network application-gateway routing-rule"):
        from .custom import RoutingRuleCreate, RoutingRuleUpdate
        self.command_table["network application-gateway routing-rule create"] = RoutingRuleCreate(loader=self)
        self.command_table["network application-gateway routing-rule update"] = RoutingRuleUpdate(loader=self)

    with self.command_group("network application-gateway rule"):
        from .custom import RuleCreate, RuleUpdate
        self.command_table["network application-gateway rule create"] = RuleCreate(loader=self)
        self.command_table["network application-gateway rule update"] = RuleUpdate(loader=self)

    with self.command_group("network application-gateway ssl-cert"):
        from .custom import SSLCertCreate, SSLCertUpdate
        self.command_table["network application-gateway ssl-cert create"] = SSLCertCreate(loader=self)
        self.command_table["network application-gateway ssl-cert update"] = SSLCertUpdate(loader=self)

    with self.command_group("network application-gateway ssl-policy"):
        from .custom import SSLPolicySet
        self.command_table["network application-gateway ssl-policy set"] = SSLPolicySet(loader=self)

    with self.command_group("network application-gateway ssl-profile"):
        from .custom import SSLProfileAdd, SSLProfileUpdate
        self.command_table["network application-gateway ssl-profile add"] = SSLProfileAdd(loader=self)
        self.command_table["network application-gateway ssl-profile update"] = SSLProfileUpdate(loader=self)

    with self.command_group("network application-gateway url-path-map"):
        from .custom import URLPathMapCreate, URLPathMapUpdate, URLPathMapRuleCreate
        self.command_table["network application-gateway url-path-map create"] = URLPathMapCreate(loader=self)
        self.command_table["network application-gateway url-path-map update"] = URLPathMapUpdate(loader=self)
        self.command_table["network application-gateway url-path-map rule create"] = URLPathMapRuleCreate(loader=self)

    with self.command_group("network application-gateway waf-config") as g:
        g.custom_command("list-rule-sets", "list_ag_waf_rule_sets", table_transformer=transform_waf_rule_sets_table_output)
        g.custom_command("set", "set_ag_waf_config", supports_no_wait=True)
        g.custom_show_command("show", "show_ag_waf_config")
    # endregion

    # region ApplicationGatewayWAFPolicy
    with self.command_group("network application-gateway waf-policy"):
        from .custom import WAFCreate
        self.command_table["network application-gateway waf-policy create"] = WAFCreate(loader=self)

    with self.command_group("network application-gateway waf-policy custom-rule match-condition"):
        from .custom import WAFCustomRuleMatchConditionAdd
        self.command_table["network application-gateway waf-policy custom-rule match-condition add"] = WAFCustomRuleMatchConditionAdd(loader=self)

    with self.command_group("network application-gateway waf-policy managed-rule exception") as g:
        g.custom_command("remove", "remove_waf_managed_rule_exception")
        g.custom_command("list", "list_waf_managed_rules")

    with self.command_group("network application-gateway waf-policy managed-rule exclusion") as g:
        g.custom_command("remove", "remove_waf_managed_rule_exclusion")
        g.custom_command("list", "list_waf_managed_rules")

    with self.command_group("network application-gateway waf-policy managed-rule exclusion rule-set") as g:
        g.custom_command("add", "add_waf_exclusion_rule_set")
        g.custom_command("remove", "remove_waf_exclusion_rule_set")
        g.custom_command("list", "list_waf_managed_rules")

    with self.command_group("network application-gateway waf-policy managed-rule rule-set") as g:
        g.custom_command("add", "add_waf_managed_rule_set")
        g.custom_command("remove", "remove_waf_managed_rule_set")
        g.custom_command("list", "list_waf_managed_rules")
        g.custom_command("update", "update_waf_managed_rule_set", validator=process_appgw_waf_policy_update)

    with self.command_group("network application-gateway waf-policy policy-setting") as g:
        from .custom import WAFPolicySettingUpdate
        self.command_table["network application-gateway waf-policy policy-setting update"] = WAFPolicySettingUpdate(loader=self)
        g.custom_command("list", "list_waf_policy_setting")
    # endregion

    # region DdosProtectionPlans
    with self.command_group('network ddos-protection') as g:
        g.custom_command('create', 'create_ddos_plan')
        g.custom_command('update', 'update_ddos_plan')
    # endregion

    # region DNS
    from .operations.dns import DNSListReferences
    self.command_table["network dns list-references"] = DNSListReferences(loader=self)

    with self.command_group('network dns zone') as g:
        g.custom_command('import', 'import_zone')
        g.custom_command('export', 'export_zone')
        g.custom_command('create', 'create_dns_zone', table_transformer=transform_dns_zone_table_output)

    supported_records = ['a', 'aaaa', 'ds', 'mx', 'naptr', 'ns', 'ptr', 'srv', 'tlsa', 'txt', 'caa']
    experimental_records = ['ds', 'naptr', 'tlsa']
    for record in supported_records:
        is_experimental = record in experimental_records
        with self.command_group('network dns record-set {}'.format(record), is_experimental=is_experimental) as g:
            g.custom_command('add-record', 'add_dns_{}_record'.format(record), transform=transform_dns_record_set_output)
            g.custom_command('remove-record', 'remove_dns_{}_record'.format(record), transform=transform_dns_record_set_output)

    with self.command_group('network dns record-set soa') as g:
        g.custom_command('update', 'update_dns_soa_record', transform=transform_dns_record_set_output)

    with self.command_group('network dns record-set cname') as g:
        g.custom_command('set-record', 'add_dns_cname_record', transform=transform_dns_record_set_output)
        g.custom_command('remove-record', 'remove_dns_cname_record', transform=transform_dns_record_set_output)

    from .operations.dns import RecordSetAShow as DNSRecordSetAShow, RecordSetAAAAShow as DNSRecordSetAAAAShow, \
        RecordSetDSShow as DNSRecordSetDSShow, RecordSetMXShow as DNSRecordSetMXShow, \
        RecordSetNAPTRShow as DNSRecordSetNAPTRShow, RecordSetNSShow as DNSRecordSetNSShow, \
        RecordSetPTRShow as DNSRecordSetPTRShow, RecordSetSRVShow as DNSRecordSetSRVShow, \
        RecordSetTLSAShow as DNSRecordSetTLSAShow, RecordSetTXTShow as DNSRecordSetTXTShow, \
        RecordSetCAAShow as DNSRecordSetCAAShow, RecordSetCNAMEShow as DNSRecordSetCNAMEShow, \
        RecordSetSOAShow as DNSRecordSetSOAShow
    self.command_table["network dns record-set a show"] = DNSRecordSetAShow(loader=self)
    self.command_table["network dns record-set aaaa show"] = DNSRecordSetAAAAShow(loader=self)
    self.command_table["network dns record-set ds show"] = DNSRecordSetDSShow(loader=self)
    self.command_table["network dns record-set mx show"] = DNSRecordSetMXShow(loader=self)
    self.command_table["network dns record-set naptr show"] = DNSRecordSetNAPTRShow(loader=self)
    self.command_table["network dns record-set ns show"] = DNSRecordSetNSShow(loader=self)
    self.command_table["network dns record-set ptr show"] = DNSRecordSetPTRShow(loader=self)
    self.command_table["network dns record-set srv show"] = DNSRecordSetSRVShow(loader=self)
    self.command_table["network dns record-set tlsa show"] = DNSRecordSetTLSAShow(loader=self)
    self.command_table["network dns record-set txt show"] = DNSRecordSetTXTShow(loader=self)
    self.command_table["network dns record-set caa show"] = DNSRecordSetCAAShow(loader=self)
    self.command_table["network dns record-set cname show"] = DNSRecordSetCNAMEShow(loader=self)
    self.command_table["network dns record-set soa show"] = DNSRecordSetSOAShow(loader=self)

    from .operations.dns import RecordSetAList as DNSRecordSetAList, RecordSetAAAAList as DNSRecordSetAAAAList, \
        RecordSetDSList as DNSRecordSetDSList, RecordSetMXList as DNSRecordSetMXList, \
        RecordSetNAPTRList as DNSRecordSetNAPTRList, RecordSetNSList as DNSRecordSetNSList, \
        RecordSetPTRList as DNSRecordSetPTRList, RecordSetSRVList as DNSRecordSetSRVList, \
        RecordSetTLSAList as DNSRecordSetTLSAList, RecordSetTXTList as DNSRecordSetTXTList, \
        RecordSetCAAList as DNSRecordSetCAAList, RecordSetCNAMEList as DNSRecordSetCNAMEList
    self.command_table["network dns record-set a list"] = DNSRecordSetAList(loader=self)
    self.command_table["network dns record-set aaaa list"] = DNSRecordSetAAAAList(loader=self)
    self.command_table["network dns record-set ds list"] = DNSRecordSetDSList(loader=self)
    self.command_table["network dns record-set mx list"] = DNSRecordSetMXList(loader=self)
    self.command_table["network dns record-set naptr list"] = DNSRecordSetNAPTRList(loader=self)
    self.command_table["network dns record-set ns list"] = DNSRecordSetNSList(loader=self)
    self.command_table["network dns record-set ptr list"] = DNSRecordSetPTRList(loader=self)
    self.command_table["network dns record-set srv list"] = DNSRecordSetSRVList(loader=self)
    self.command_table["network dns record-set tlsa list"] = DNSRecordSetTLSAList(loader=self)
    self.command_table["network dns record-set txt list"] = DNSRecordSetTXTList(loader=self)
    self.command_table["network dns record-set caa list"] = DNSRecordSetCAAList(loader=self)
    self.command_table["network dns record-set cname list"] = DNSRecordSetCNAMEList(loader=self)

    from .operations.dns import RecordSetACreate as DNSRecordSetACreate, RecordSetAAAACreate as DNSRecordSetAAAACreate, \
        RecordSetDSCreate as DNSRecordSetDSCreate, RecordSetMXCreate as DNSRecordSetMXCreate, \
        RecordSetNAPTRCreate as DNSRecordSetNAPTRCreate, RecordSetNSCreate as DNSRecordSetNSCreate, \
        RecordSetPTRCreate as DNSRecordSetPTRCreate, RecordSetSRVCreate as DNSRecordSetSRVCreate, \
        RecordSetTLSACreate as DNSRecordSetTLSACreate, RecordSetTXTCreate as DNSRecordSetTXTCreate, \
        RecordSetCAACreate as DNSRecordSetCAACreate, RecordSetCNAMECreate as DNSRecordSetCNAMECreate
    self.command_table["network dns record-set a create"] = DNSRecordSetACreate(loader=self)
    self.command_table["network dns record-set aaaa create"] = DNSRecordSetAAAACreate(loader=self)
    self.command_table["network dns record-set ds create"] = DNSRecordSetDSCreate(loader=self)
    self.command_table["network dns record-set mx create"] = DNSRecordSetMXCreate(loader=self)
    self.command_table["network dns record-set naptr create"] = DNSRecordSetNAPTRCreate(loader=self)
    self.command_table["network dns record-set ns create"] = DNSRecordSetNSCreate(loader=self)
    self.command_table["network dns record-set ptr create"] = DNSRecordSetPTRCreate(loader=self)
    self.command_table["network dns record-set srv create"] = DNSRecordSetSRVCreate(loader=self)
    self.command_table["network dns record-set tlsa create"] = DNSRecordSetTLSACreate(loader=self)
    self.command_table["network dns record-set txt create"] = DNSRecordSetTXTCreate(loader=self)
    self.command_table["network dns record-set caa create"] = DNSRecordSetCAACreate(loader=self)
    self.command_table["network dns record-set cname create"] = DNSRecordSetCNAMECreate(loader=self)

    from .operations.dns import RecordSetAUpdate as DNSRecordSetAUpdate, RecordSetAAAAUpdate as DNSRecordSetAAAAUpdate, \
        RecordSetDSUpdate as DNSRecordSetDSUpdate, RecordSetMXUpdate as DNSRecordSetMXUpdate, \
        RecordSetNAPTRUpdate as DNSRecordSetNAPTRUpdate, RecordSetNSUpdate as DNSRecordSetNSUpdate, \
        RecordSetPTRUpdate as DNSRecordSetPTRUpdate, RecordSetSRVUpdate as DNSRecordSetSRVUpdate, \
        RecordSetTLSAUpdate as DNSRecordSetTLSAUpdate, RecordSetTXTUpdate as DNSRecordSetTXTUpdate, \
        RecordSetCAAUpdate as DNSRecordSetCAAUpdate, RecordSetCNAMEUpdate as DNSRecordSetCNAMEUpdate
    self.command_table["network dns record-set a update"] = DNSRecordSetAUpdate(loader=self)
    self.command_table["network dns record-set aaaa update"] = DNSRecordSetAAAAUpdate(loader=self)
    self.command_table["network dns record-set ds update"] = DNSRecordSetDSUpdate(loader=self)
    self.command_table["network dns record-set mx update"] = DNSRecordSetMXUpdate(loader=self)
    self.command_table["network dns record-set naptr update"] = DNSRecordSetNAPTRUpdate(loader=self)
    self.command_table["network dns record-set ns update"] = DNSRecordSetNSUpdate(loader=self)
    self.command_table["network dns record-set ptr update"] = DNSRecordSetPTRUpdate(loader=self)
    self.command_table["network dns record-set srv update"] = DNSRecordSetSRVUpdate(loader=self)
    self.command_table["network dns record-set tlsa update"] = DNSRecordSetTLSAUpdate(loader=self)
    self.command_table["network dns record-set txt update"] = DNSRecordSetTXTUpdate(loader=self)
    self.command_table["network dns record-set caa update"] = DNSRecordSetCAAUpdate(loader=self)
    self.command_table["network dns record-set cname update"] = DNSRecordSetCNAMEUpdate(loader=self)

    from .operations.dns import RecordSetADelete as DNSRecordSetADelete, RecordSetAAAADelete as DNSRecordSetAAAADelete, \
        RecordSetDSDelete as DNSRecordSetDSDelete, RecordSetMXDelete as DNSRecordSetMXDelete, \
        RecordSetNAPTRDelete as DNSRecordSetNAPTRDelete, RecordSetNSDelete as DNSRecordSetNSDelete, \
        RecordSetPTRDelete as DNSRecordSetPTRDelete, RecordSetSRVDelete as DNSRecordSetSRVDelete, \
        RecordSetTLSADelete as DNSRecordSetTLSADelete, RecordSetTXTDelete as DNSRecordSetTXTDelete, \
        RecordSetCAADelete as DNSRecordSetCAADelete, RecordSetCNAMEDelete as DNSRecordSetCNAMEDelete
    self.command_table["network dns record-set a delete"] = DNSRecordSetADelete(loader=self)
    self.command_table["network dns record-set aaaa delete"] = DNSRecordSetAAAADelete(loader=self)
    self.command_table["network dns record-set ds delete"] = DNSRecordSetDSDelete(loader=self)
    self.command_table["network dns record-set mx delete"] = DNSRecordSetMXDelete(loader=self)
    self.command_table["network dns record-set naptr delete"] = DNSRecordSetNAPTRDelete(loader=self)
    self.command_table["network dns record-set ns delete"] = DNSRecordSetNSDelete(loader=self)
    self.command_table["network dns record-set ptr delete"] = DNSRecordSetPTRDelete(loader=self)
    self.command_table["network dns record-set srv delete"] = DNSRecordSetSRVDelete(loader=self)
    self.command_table["network dns record-set tlsa delete"] = DNSRecordSetTLSADelete(loader=self)
    self.command_table["network dns record-set txt delete"] = DNSRecordSetTXTDelete(loader=self)
    self.command_table["network dns record-set caa delete"] = DNSRecordSetCAADelete(loader=self)
    self.command_table["network dns record-set cname delete"] = DNSRecordSetCNAMEDelete(loader=self)
    # endregion

    # region ExpressRoutes
    with self.command_group('network express-route'):
        from .custom import ExpressRouteCreate, ExpressRouteUpdate
        self.command_table['network express-route create'] = ExpressRouteCreate(loader=self)
        self.command_table['network express-route update'] = ExpressRouteUpdate(loader=self)

    with self.command_group('network express-route gateway'):
        from .custom import ExpressRouteGatewayCreate, ExpressRouteGatewayUpdate
        self.command_table['network express-route gateway create'] = ExpressRouteGatewayCreate(loader=self)
        self.command_table['network express-route gateway update'] = ExpressRouteGatewayUpdate(loader=self)

    with self.command_group('network express-route gateway connection'):
        from .custom import ExpressRouteConnectionUpdate, ExpressRouteConnectionCreate
        self.command_table['network express-route gateway connection create'] = ExpressRouteConnectionCreate(loader=self)
        self.command_table['network express-route gateway connection update'] = ExpressRouteConnectionUpdate(loader=self)

    with self.command_group('network express-route peering'):
        from .custom import ExpressRoutePeeringCreate, ExpressRoutePeeringUpdate
        self.command_table['network express-route peering create'] = ExpressRoutePeeringCreate(loader=self)
        self.command_table['network express-route peering update'] = ExpressRoutePeeringUpdate(loader=self)

    with self.command_group('network express-route peering connection'):
        from .custom import ExpressRoutePeeringConnectionCreate
        self.command_table['network express-route peering connection create'] = ExpressRoutePeeringConnectionCreate(loader=self)

    with self.command_group('network express-route port') as g:
        from .custom import ExpressRoutePortCreate
        self.command_table['network express-route port create'] = ExpressRoutePortCreate(loader=self)
        g.custom_command('generate-loa', 'download_generated_loa_as_pdf')

    with self.command_group('network express-route port identity'):
        from .custom import ExpressRoutePortIdentityAssign
        self.command_table['network express-route port identity assign'] = ExpressRoutePortIdentityAssign(loader=self)

    with self.command_group('network express-route port link'):
        from .custom import ExpressRoutePortLinkUpdate
        self.command_table['network express-route port link update'] = ExpressRoutePortLinkUpdate(loader=self)
    # endregion

    # region PrivateEndpoint
    with self.command_group('network private-endpoint'):
        from azure.cli.command_modules.network.custom import PrivateEndpointCreate, PrivateEndpointUpdate
        self.command_table['network private-endpoint create'] = PrivateEndpointCreate(loader=self)
        self.command_table['network private-endpoint update'] = PrivateEndpointUpdate(loader=self)

    with self.command_group('network private-endpoint dns-zone-group'):
        from azure.cli.command_modules.network.custom import PrivateEndpointPrivateDnsZoneGroupCreate, \
            PrivateEndpointPrivateDnsZoneAdd
        self.command_table['network private-endpoint dns-zone-group create'] = \
            PrivateEndpointPrivateDnsZoneGroupCreate(loader=self)
        self.command_table['network private-endpoint dns-zone-group add'] = \
            PrivateEndpointPrivateDnsZoneAdd(loader=self)

    with self.command_group('network private-endpoint ip-config'):
        from azure.cli.command_modules.network.custom import PrivateEndpointIpConfigAdd
        self.command_table['network private-endpoint ip-config add'] = PrivateEndpointIpConfigAdd(loader=self)

    with self.command_group('network private-endpoint asg'):
        from azure.cli.command_modules.network.custom import PrivateEndpointAsgAdd
        self.command_table['network private-endpoint asg add'] = PrivateEndpointAsgAdd(loader=self)
    # endregion

    # region PrivateLinkServices
    with self.command_group('network private-link-service'):
        from azure.cli.command_modules.network.custom import PrivateLinkServiceCreate, PrivateLinkServiceUpdate
        self.command_table['network private-link-service create'] = PrivateLinkServiceCreate(loader=self)
        self.command_table['network private-link-service update'] = PrivateLinkServiceUpdate(loader=self)

    with self.command_group('network private-link-service connection'):
        from azure.cli.command_modules.network.custom import PrivateEndpointConnectionUpdate
        self.command_table['network private-link-service connection update'] = PrivateEndpointConnectionUpdate(loader=self)
    # endregion

    # region LoadBalancers
    with self.command_group('network lb') as g:
        g.custom_command('create', 'create_load_balancer',
                         transform=DeploymentOutputLongRunningOperation(self.cli_ctx),
                         supports_no_wait=True,
                         table_transformer=deployment_validate_table_format,
                         validator=process_lb_create_namespace,
                         exception_handler=handle_template_based_exception)
        g.custom_command('list-mapping', 'list_load_balancer_mapping')

    from .operations.load_balancer import LBFrontendIPCreate, LBFrontendIPUpdate
    self.command_table["network lb frontend-ip create"] = LBFrontendIPCreate(loader=self)
    self.command_table["network lb frontend-ip update"] = LBFrontendIPUpdate(loader=self)

    from .operations.load_balancer import LBInboundNatRuleCreate, LBInboundNatRuleUpdate
    self.command_table["network lb inbound-nat-rule create"] = LBInboundNatRuleCreate(loader=self)
    self.command_table["network lb inbound-nat-rule update"] = LBInboundNatRuleUpdate(loader=self)

    from .operations.load_balancer import LBInboundNatPoolCreate, LBInboundNatPoolUpdate
    self.command_table["network lb inbound-nat-pool create"] = LBInboundNatPoolCreate(loader=self)
    self.command_table["network lb inbound-nat-pool update"] = LBInboundNatPoolUpdate(loader=self)

    from .operations.load_balancer import LBRuleCreate, LBRuleUpdate
    self.command_table["network lb rule create"] = LBRuleCreate(loader=self)
    self.command_table["network lb rule update"] = LBRuleUpdate(loader=self)

    from .operations.load_balancer import LBOutboundRuleCreate, LBOutboundRuleUpdate
    self.command_table["network lb outbound-rule create"] = LBOutboundRuleCreate(loader=self)
    self.command_table["network lb outbound-rule update"] = LBOutboundRuleUpdate(loader=self)

    from .operations.load_balancer import LBAddressPoolCreate, LBAddressPoolDelete, LBAddressPoolUpdate
    self.command_table["network lb address-pool create"] = LBAddressPoolCreate(loader=self)
    self.command_table["network lb address-pool delete"] = LBAddressPoolDelete(loader=self)
    self.command_table["network lb address-pool update"] = LBAddressPoolUpdate(loader=self)

    from .operations.load_balancer import LBAddressPoolAddressAdd, LBAddressPoolAddressUpdate
    self.command_table["network lb address-pool address add"] = LBAddressPoolAddressAdd(loader=self)
    self.command_table["network lb address-pool address update"] = LBAddressPoolAddressUpdate(loader=self)

    from .operations.load_balancer import LBAddressPoolTunnelInterfaceAdd, LBAddressPoolTunnelInterfaceUpdate
    self.command_table["network lb address-pool tunnel-interface add"] = LBAddressPoolTunnelInterfaceAdd(loader=self)
    self.command_table["network lb address-pool tunnel-interface update"] = LBAddressPoolTunnelInterfaceUpdate(loader=self)

    from .operations.load_balancer import LBProbeCreate, LBProbeUpdate
    self.command_table["network lb probe create"] = LBProbeCreate(loader=self)
    self.command_table["network lb probe update"] = LBProbeUpdate(loader=self)

    # endregion

    # region cross-region load balancer
    with self.command_group('network cross-region-lb') as g:
        g.custom_command('create', 'create_cross_region_load_balancer', transform=DeploymentOutputLongRunningOperation(self.cli_ctx), supports_no_wait=True, table_transformer=deployment_validate_table_format, validator=process_cross_region_lb_create_namespace, exception_handler=handle_template_based_exception)

        from .aaz.latest.network.lb import Wait
        from .operations.load_balancer import CrossRegionLoadBalancerShow, CrossRegionLoadBalancerDelete, CrossRegionLoadBalancerUpdate, CrossRegionLoadBalancerList
        self.command_table['network cross-region-lb show'] = CrossRegionLoadBalancerShow(loader=self)
        self.command_table['network cross-region-lb delete'] = CrossRegionLoadBalancerDelete(loader=self)
        self.command_table['network cross-region-lb list'] = CrossRegionLoadBalancerList(loader=self)
        self.command_table['network cross-region-lb update'] = CrossRegionLoadBalancerUpdate(loader=self)
        self.command_table['network cross-region-lb wait'] = Wait(loader=self)

    with self.command_group('network cross-region-lb frontend-ip') as g:
        from .operations.load_balancer import CrossRegionLoadBalancerFrontendIPShow, CrossRegionLoadBalancerFrontendIPList, CrossRegionLoadBalancerFrontendIPDelete, CrossRegionLoadBalancerFrontendIPCreate, CrossRegionLoadBalancerFrontendIPUpdate
        self.command_table['network cross-region-lb frontend-ip show'] = CrossRegionLoadBalancerFrontendIPShow(loader=self)
        self.command_table['network cross-region-lb frontend-ip delete'] = CrossRegionLoadBalancerFrontendIPDelete(loader=self)
        self.command_table['network cross-region-lb frontend-ip list'] = CrossRegionLoadBalancerFrontendIPList(loader=self)
        self.command_table['network cross-region-lb frontend-ip create'] = CrossRegionLoadBalancerFrontendIPCreate(loader=self)
        self.command_table['network cross-region-lb frontend-ip update'] = CrossRegionLoadBalancerFrontendIPUpdate(loader=self)

    with self.command_group('network cross-region-lb rule') as g:
        from .operations.load_balancer import CrossRegionLoadBalancerRuleShow, CrossRegionLoadBalancerRuleDelete, CrossRegionLoadBalancerRuleList, CrossRegionLoadBalancerRuleCreate, CrossRegionLoadBalancerRuleUpdate
        self.command_table['network cross-region-lb rule show'] = CrossRegionLoadBalancerRuleShow(loader=self)
        self.command_table['network cross-region-lb rule delete'] = CrossRegionLoadBalancerRuleDelete(loader=self)
        self.command_table['network cross-region-lb rule list'] = CrossRegionLoadBalancerRuleList(loader=self)
        self.command_table['network cross-region-lb rule create'] = CrossRegionLoadBalancerRuleCreate(loader=self)
        self.command_table['network cross-region-lb rule update'] = CrossRegionLoadBalancerRuleUpdate(loader=self)

    with self.command_group('network cross-region-lb address-pool') as g:
        from .operations.load_balancer import CrossRegionLoadBalancerAddressPoolCreate, CrossRegionLoadBalancerAddressPoolUpdate, CrossRegionLoadBalancerAddressPoolList, CrossRegionLoadBalancerAddressPoolDelete, CrossRegionLoadBalancerAddressPoolShow
        self.command_table['network cross-region-lb address-pool show'] = CrossRegionLoadBalancerAddressPoolShow(loader=self)
        self.command_table['network cross-region-lb address-pool delete'] = CrossRegionLoadBalancerAddressPoolDelete(loader=self)
        self.command_table['network cross-region-lb address-pool list'] = CrossRegionLoadBalancerAddressPoolList(loader=self)
        self.command_table['network cross-region-lb address-pool create'] = CrossRegionLoadBalancerAddressPoolCreate(loader=self)
        self.command_table['network cross-region-lb address-pool update'] = CrossRegionLoadBalancerAddressPoolUpdate(loader=self)

    with self.command_group('network cross-region-lb address-pool address') as g:
        from .operations.load_balancer import CrossRegionLoadBalancerAddressPoolAddressAdd, CrossRegionLoadBalancerAddressPoolAddressUpdate, CrossRegionLoadBalancerAddressPoolAddressRemove, CrossRegionLoadBalancerAddressPoolAddressList, CrossRegionLoadBalancerAddressPoolAddressShow
        self.command_table['network cross-region-lb address-pool address add'] = CrossRegionLoadBalancerAddressPoolAddressAdd(loader=self)
        self.command_table['network cross-region-lb address-pool address remove'] = CrossRegionLoadBalancerAddressPoolAddressRemove(loader=self)
        self.command_table['network cross-region-lb address-pool address update'] = CrossRegionLoadBalancerAddressPoolAddressUpdate(loader=self)
        self.command_table['network cross-region-lb address-pool address list'] = CrossRegionLoadBalancerAddressPoolAddressList(loader=self)
        self.command_table['network cross-region-lb address-pool address show'] = CrossRegionLoadBalancerAddressPoolAddressShow(loader=self)
    # endregion

    # region LocalGateways
    with self.command_group('network local-gateway'):
        from .aaz.latest.network.local_gateway import List
        self.command_table['network local-gateway list'] = List(loader=self, table_transformer=transform_local_gateway_table_output)
    # endregion

    # region NetworkInterfaces: (NIC)
    with self.command_group("network nic"):
        from .aaz.latest.network.nic import ListEffectiveNsg, ShowEffectiveRouteTable
        from .custom import NICCreate, NICUpdate
        self.command_table["network nic create"] = NICCreate(loader=self)
        self.command_table["network nic update"] = NICUpdate(loader=self)
        self.command_table["network nic list-effective-nsg"] = ListEffectiveNsg(loader=self, table_transformer=transform_effective_nsg)
        self.command_table["network nic show-effective-route-table"] = ShowEffectiveRouteTable(loader=self, table_transformer=transform_effective_route_table)

    with self.command_group("network nic ip-config"):
        from .custom import NICIPConfigCreate, NICIPConfigUpdate, NICIPConfigNATAdd, NICIPConfigNATRemove
        self.command_table["network nic ip-config create"] = NICIPConfigCreate(loader=self)
        self.command_table["network nic ip-config update"] = NICIPConfigUpdate(loader=self)
        self.command_table["network nic ip-config inbound-nat-rule add"] = NICIPConfigNATAdd(loader=self)
        self.command_table["network nic ip-config inbound-nat-rule remove"] = NICIPConfigNATRemove(loader=self)

    with self.command_group("network nic ip-config address-pool") as g:
        g.custom_command("add", "add_nic_ip_config_address_pool")
        g.custom_command("remove", "remove_nic_ip_config_address_pool")
    # endregion

    # region NetworkSecurityGroups
    with self.command_group("network nsg") as g:
        from .custom import NSGCreate
        self.command_table["network nsg create"] = NSGCreate(loader=self)

    with self.command_group("network nsg rule") as g:
        from .custom import NSGRuleCreate, NSGRuleUpdate
        from .aaz.latest.network.nsg.rule import Show
        self.command_table["network nsg rule create"] = NSGRuleCreate(loader=self)
        self.command_table["network nsg rule update"] = NSGRuleUpdate(loader=self)
        self.command_table["network nsg rule show"] = Show(loader=self, table_transformer=transform_nsg_rule_table_output)
        g.custom_command("list", "list_nsg_rules", table_transformer=lambda x: [transform_nsg_rule_table_output(i) for i in x])
    # endregion

    # region NetworkWatchers
    with self.command_group("network watcher") as g:
        from .operations.watcher import RunConfigurationDiagnostic, ShowNextHop, ShowSecurityGroupView, ShowTopology, TestConnectivity, TestIPFlow
        self.command_table["network watcher test-ip-flow"] = TestIPFlow(loader=self)
        self.command_table["network watcher show-next-hop"] = ShowNextHop(loader=self)
        self.command_table["network watcher show-security-group-view"] = ShowSecurityGroupView(loader=self)
        self.command_table["network watcher run-configuration-diagnostic"] = RunConfigurationDiagnostic(loader=self)
        self.command_table["network watcher show-topology"] = ShowTopology(loader=self)
        self.command_table["network watcher test-connectivity"] = TestConnectivity(loader=self)
        g.custom_command("configure", "configure_network_watcher")

    from .operations.watcher import WatcherConnectionMonitorStart, WatcherConnectionMonitorStop, \
        WatcherConnectionMonitorList, WatcherConnectionMonitorQuery, WatcherConnectionMonitorShow, \
        WatcherConnectionMonitorDelete, WatcherConnectionMonitorCreate
    self.command_table["network watcher connection-monitor create"] = WatcherConnectionMonitorCreate(loader=self)
    self.command_table["network watcher connection-monitor start"] = WatcherConnectionMonitorStart(loader=self)
    self.command_table["network watcher connection-monitor stop"] = WatcherConnectionMonitorStop(loader=self)
    self.command_table["network watcher connection-monitor list"] = WatcherConnectionMonitorList(loader=self)
    self.command_table["network watcher connection-monitor query"] = WatcherConnectionMonitorQuery(loader=self)
    self.command_table["network watcher connection-monitor delete"] = WatcherConnectionMonitorDelete(loader=self)
    self.command_table["network watcher connection-monitor show"] = WatcherConnectionMonitorShow(loader=self)

    from .operations.watcher import WatcherConnectionMonitorEndpointAdd, WatcherConnectionMonitorEndpointShow, \
        WatcherConnectionMonitorEndpointList, WatcherConnectionMonitorEndpointRemove
    self.command_table["network watcher connection-monitor endpoint add"] = WatcherConnectionMonitorEndpointAdd(
        loader=self)
    self.command_table["network watcher connection-monitor endpoint remove"] = WatcherConnectionMonitorEndpointRemove(
        loader=self)
    self.command_table["network watcher connection-monitor endpoint show"] = WatcherConnectionMonitorEndpointShow(
        loader=self)
    self.command_table["network watcher connection-monitor endpoint list"] = WatcherConnectionMonitorEndpointList(
        loader=self)

    from .operations.watcher import WatcherConnectionMonitorTestConfigurationAdd as WCMTAdd, \
        WatcherConnectionMonitorTestConfigurationShow as WCMTCShow, \
        WatcherConnectionMonitorTestConfigurationList as WCMTCList, \
        WatcherConnectionMonitorTestConfigurationRemove as WCMTRemove
    self.command_table["network watcher connection-monitor test-configuration add"] = WCMTAdd(loader=self)
    self.command_table["network watcher connection-monitor test-configuration remove"] = WCMTRemove(loader=self)
    self.command_table["network watcher connection-monitor test-configuration show"] = WCMTCShow(loader=self)
    self.command_table["network watcher connection-monitor test-configuration list"] = WCMTCList(loader=self)

    with self.command_group('network watcher connection-monitor test-group', is_preview=True) as c:
        from .operations.watcher import WatcherConnectionMonitorTestGroupAdd as WCMTGAdd, \
            WatcherConnectionMonitorTestGroupShow as WCMTGShow, \
            WatcherConnectionMonitorTestGroupList as WCMTGList
        self.command_table["network watcher connection-monitor test-group add"] = WCMTGAdd(loader=self)
        self.command_table["network watcher connection-monitor test-group show"] = WCMTGShow(loader=self)
        self.command_table["network watcher connection-monitor test-group list"] = WCMTGList(loader=self)
        c.custom_command('remove', 'remove_nw_connection_monitor_test_group')

    with self.command_group('network watcher connection-monitor output', is_preview=True) as c:
        from .operations.watcher import WatcherConnectionMonitorOutputAdd, WatcherConnectionMonitorOutputList
        self.command_table["network watcher connection-monitor output add"] = WatcherConnectionMonitorOutputAdd(
            loader=self)
        self.command_table["network watcher connection-monitor output list"] = WatcherConnectionMonitorOutputList(
            loader=self)
        c.custom_command('remove', 'remove_nw_connection_monitor_output')

    with self.command_group("network watcher packet-capture"):
        from .operations.watcher import PacketCaptureCreate, PacketCaptureDelete, PacketCaptureList, PacketCaptureShow, PacketCaptureShowStatus, PacketCaptureStop
        self.command_table["network watcher packet-capture create"] = PacketCaptureCreate(loader=self)
        self.command_table["network watcher packet-capture delete"] = PacketCaptureDelete(loader=self)
        self.command_table["network watcher packet-capture list"] = PacketCaptureList(loader=self)
        self.command_table["network watcher packet-capture show"] = PacketCaptureShow(loader=self)
        self.command_table["network watcher packet-capture show-status"] = PacketCaptureShowStatus(loader=self)
        self.command_table["network watcher packet-capture stop"] = PacketCaptureStop(loader=self)

    with self.command_group('network watcher flow-log') as g:
        from .operations.watcher import NwFlowLogCreate, NwFlowLogUpdate, NwFlowLogList, NwFlowLogDelete
        self.command_table["network watcher flow-log create"] = NwFlowLogCreate(loader=self)
        self.command_table["network watcher flow-log update"] = NwFlowLogUpdate(loader=self)
        self.command_table["network watcher flow-log list"] = NwFlowLogList(loader=self)
        self.command_table["network watcher flow-log delete"] = NwFlowLogDelete(loader=self)
        g.custom_show_command('show', 'show_nw_flow_logging', validator=process_nw_flow_log_show_namespace)

    with self.command_group('network watcher troubleshooting'):
        from .operations.watcher import NwTroubleshootingStart, NwTroubleshootingShow
        self.command_table["network watcher troubleshooting start"] = NwTroubleshootingStart(loader=self)
        self.command_table["network watcher troubleshooting show"] = NwTroubleshootingShow(loader=self)
    # endregion

    # region PublicIPAddresses
    public_ip_show_table_transform = '{Name:name, ResourceGroup:resourceGroup, Location:location, $zone$Address:ipAddress, AddressVersion:publicIpAddressVersion, AllocationMethod:publicIpAllocationMethod, IdleTimeoutInMinutes:idleTimeoutInMinutes, ProvisioningState:provisioningState}'
    public_ip_show_table_transform = public_ip_show_table_transform.replace('$zone$', 'Zones: (!zones && \' \') || join(` `, zones), ')

    with self.command_group('network public-ip') as g:
        from .aaz.latest.network.public_ip import List, Show
        from .custom import PublicIPUpdate
        self.command_table['network public-ip update'] = PublicIPUpdate(loader=self)
        self.command_table['network public-ip list'] = List(loader=self, table_transformer='[].' + public_ip_show_table_transform)
        self.command_table['network public-ip show'] = Show(loader=self, table_transformer=public_ip_show_table_transform)
        g.custom_command('create', 'create_public_ip', transform=transform_public_ip_create_output, validator=process_public_ip_create_namespace)

    with self.command_group('network public-ip prefix'):
        from azure.cli.command_modules.network.custom import PublicIpPrefixCreate
        self.command_table['network public-ip prefix create'] = PublicIpPrefixCreate(loader=self)
    # endregion

    # region RouteFilters
    from azure.cli.command_modules.network.aaz.latest.network.route_filter.rule import ListServiceCommunities
    self.command_table['network route-filter rule list-service-communities'] = ListServiceCommunities(loader=self, table_transformer=transform_service_community_table_output)
    # endregion

    # region TrafficManagers
    with self.command_group('network traffic-manager profile') as g:
        g.custom_command('create', 'create_traffic_manager_profile', transform=transform_traffic_manager_create_output)
        g.custom_command('update', 'update_traffic_manager_profile')

    with self.command_group('network traffic-manager endpoint') as g:
        g.custom_command('create', 'create_traffic_manager_endpoint')
        g.custom_command('update', 'update_traffic_manager_endpoint')
        g.custom_command('list', 'list_traffic_manager_endpoints')

    from azure.cli.command_modules.network.aaz.latest.network.traffic_manager.endpoint import ShowGeographicHierarchy
    self.command_table['network traffic-manager endpoint show-geographic-hierarchy'] = ShowGeographicHierarchy(loader=self, table_transformer=transform_geographic_hierachy_table_output)
    # endregion

    # region VirtualNetworks
    with self.command_group("network vnet") as g:
        from .aaz.latest.network.vnet import List
        from .custom import VNetCreate, VNetUpdate
        self.command_table["network vnet create"] = VNetCreate(loader=self)
        self.command_table["network vnet update"] = VNetUpdate(loader=self)
        self.command_table['network vnet list'] = List(loader=self, table_transformer=transform_vnet_table_output)
        g.custom_command("list-available-ips", "list_available_ips", is_preview=True)

    with self.command_group("network vnet peering") as g:
        from .custom import VNetPeeringCreate
        self.command_table["network vnet peering create"] = VNetPeeringCreate(loader=self)
        g.custom_command("sync", "sync_vnet_peering")

    with self.command_group("network vnet subnet") as g:
        from .custom import VNetSubnetCreate, VNetSubnetUpdate
        self.command_table["network vnet subnet create"] = VNetSubnetCreate(loader=self)
        self.command_table["network vnet subnet update"] = VNetSubnetUpdate(loader=self)
        g.custom_command("list-available-ips", "subnet_list_available_ips", is_preview=True)
    # endregion

    # region VirtualNetworkGateways
    with self.command_group('network vnet-gateway'):
        from .custom import VnetGatewayCreate, VnetGatewayUpdate, VnetGatewayVpnConnectionsDisconnect, VNetGatewayShow, VNetGatewayList
        from .aaz.latest.network.vnet_gateway import ListBgpPeerStatus, ListAdvertisedRoutes, ListLearnedRoutes
        self.command_table['network vnet-gateway create'] = VnetGatewayCreate(loader=self)
        self.command_table['network vnet-gateway update'] = VnetGatewayUpdate(loader=self)
        self.command_table['network vnet-gateway disconnect-vpn-connections'] = VnetGatewayVpnConnectionsDisconnect(loader=self)
        self.command_table["network vnet-gateway show"] = VNetGatewayShow(loader=self)
        self.command_table["network vnet-gateway list"] = VNetGatewayList(loader=self)
        self.command_table['network vnet-gateway list-bgp-peer-status'] = ListBgpPeerStatus(loader=self, table_transformer=transform_vnet_gateway_bgp_peer_table)
        self.command_table['network vnet-gateway list-advertised-routes'] = ListAdvertisedRoutes(loader=self, table_transformer=transform_vnet_gateway_routes_table)
        self.command_table['network vnet-gateway list-learned-routes'] = ListLearnedRoutes(loader=self, table_transformer=transform_vnet_gateway_routes_table)

    with self.command_group('network vnet-gateway migration') as g:
        from .operations.vnet_gateway_migration import VNetGatewayMigrationAbort, VNetGatewayMigrationExecute, \
            VNetGatewayMigrationCommit, VNetGatewayMigrationPrepare
        self.command_table['network vnet-gateway migration abort'] = VNetGatewayMigrationAbort(loader=self)
        self.command_table['network vnet-gateway migration execute'] = VNetGatewayMigrationExecute(loader=self)
        self.command_table['network vnet-gateway migration commit'] = VNetGatewayMigrationCommit(loader=self)
        self.command_table['network vnet-gateway migration prepare'] = VNetGatewayMigrationPrepare(loader=self)

    with self.command_group('network vnet-gateway packet-capture'):
        from .aaz.latest.network.vnet_gateway import Wait
        self.command_table['network vnet-gateway packet-capture wait'] = Wait(loader=self)

    with self.command_group('network vnet-gateway vpn-client') as g:
        g.custom_command('generate', 'generate_vpn_client')

    with self.command_group('network vnet-gateway vpn-client ipsec-policy'):
        from .aaz.latest.network.vnet_gateway import Wait
        self.command_table['network vnet-gateway vpn-client ipsec-policy wait'] = Wait(loader=self)

    with self.command_group('network vnet-gateway revoked-cert'):
        from .custom import VnetGatewayRevokedCertCreate
        self.command_table['network vnet-gateway revoked-cert create'] = VnetGatewayRevokedCertCreate(loader=self)

    with self.command_group('network vnet-gateway root-cert'):
        from .custom import VnetGatewayRootCertCreate
        self.command_table['network vnet-gateway root-cert create'] = VnetGatewayRootCertCreate(loader=self)

    with self.command_group('network vnet-gateway ipsec-policy') as g:
        from .custom import VnetGatewayIpsecPolicyAdd
        self.command_table['network vnet-gateway ipsec-policy add'] = VnetGatewayIpsecPolicyAdd(loader=self)
        g.custom_command('clear', 'clear_vnet_gateway_ipsec_policies', supports_no_wait=True)

    with self.command_group('network vnet-gateway aad') as g:
        from .custom import VnetGatewayAadAssign
        self.command_table['network vnet-gateway aad assign'] = VnetGatewayAadAssign(loader=self)
        g.custom_command('remove', 'remove_vnet_gateway_aad', supports_no_wait=True)

    with self.command_group('network vnet-gateway nat-rule'):
        from .custom import VnetGatewayNatRuleAdd, VnetGatewayNatRuleShow, VnetGatewayNatRuleRemove
        self.command_table['network vnet-gateway nat-rule add'] = VnetGatewayNatRuleAdd(loader=self)
        self.command_table['network vnet-gateway nat-rule list'] = VnetGatewayNatRuleShow(loader=self)
        self.command_table['network vnet-gateway nat-rule remove'] = VnetGatewayNatRuleRemove(loader=self)
    # endregion

    # region VirtualNetworkGatewayConnections
    with self.command_group('network vpn-connection') as g:
        from .custom import VpnConnectionUpdate, VpnConnectionDeviceConfigScriptShow
        self.command_table['network vpn-connection update'] = VpnConnectionUpdate(loader=self)
        self.command_table['network vpn-connection show-device-config-script'] = VpnConnectionDeviceConfigScriptShow(loader=self)
        g.custom_command('create', 'create_vpn_connection', transform=DeploymentOutputLongRunningOperation(self.cli_ctx), table_transformer=deployment_validate_table_format, validator=process_vpn_connection_create_namespace, exception_handler=handle_template_based_exception)
        g.custom_command('list', 'list_vpn_connections')

    with self.command_group('network vpn-connection shared-key'):
        from .custom import VpnConnSharedKeyUpdate
        self.command_table['network vpn-connection shared-key update'] = VpnConnSharedKeyUpdate(loader=self)

    with self.command_group('network vpn-connection ipsec-policy') as g:
        from .custom import VpnConnIpsecPolicyAdd
        self.command_table['network vpn-connection ipsec-policy add'] = VpnConnIpsecPolicyAdd(loader=self)
        g.custom_command('clear', 'clear_vpn_conn_ipsec_policies', supports_no_wait=True)

    with self.command_group('network vpn-connection packet-capture'):
        from .custom import VpnConnPackageCaptureStop
        from .aaz.latest.network.vpn_connection import Wait
        self.command_table['network vpn-connection packet-capture stop'] = VpnConnPackageCaptureStop(loader=self)
        self.command_table['network vpn-connection packet-capture wait'] = Wait(loader=self)
    # endregion

    # region VirtualHub
    with self.command_group('network routeserver') as g:
        g.custom_command('create', 'create_virtual_hub')
        g.custom_command('delete', 'delete_virtual_hub', supports_no_wait=True, confirmation=True)
    # endregion

    # region PrivateLinkResource and PrivateEndpointConnection
    plr_and_pec_custom = CliCommandType(operations_tmpl='azure.cli.command_modules.network.private_link_resource_and_endpoint_connections.custom#{}')
    with self.command_group('network private-link-resource', custom_command_type=plr_and_pec_custom) as g:
        g.custom_show_command('list', 'list_private_link_resource')
    with self.command_group('network private-endpoint-connection', custom_command_type=plr_and_pec_custom) as g:
        g.custom_command('approve', 'approve_private_endpoint_connection')
        g.custom_command('reject', 'reject_private_endpoint_connection')
        g.custom_command('delete', 'remove_private_endpoint_connection', confirmation=True)
        g.custom_show_command('show', 'show_private_endpoint_connection')
        g.custom_command('list', 'list_private_endpoint_connection')
    # endregion

    # region NatGateway
    from .operations.nat import GatewayCreate as NATGatewayCreate, GatewayUpdate as NATGatewayUpdate
    self.command_table["network nat gateway create"] = NATGatewayCreate(loader=self)
    self.command_table["network nat gateway update"] = NATGatewayUpdate(loader=self)
    # endregion

    # region SecurityPartnerProvider
    from .custom import SecurityPartnerProviderCreate, SecurityPartnerProviderUpdate
    self.command_table["network security-partner-provider create"] = SecurityPartnerProviderCreate(loader=self)
    self.command_table["network security-partner-provider update"] = SecurityPartnerProviderUpdate(loader=self)
    # endregion

    # region VirtualAppliance
    from .custom import VirtualApplianceCreate, VirtualApplianceUpdate
    self.command_table["network virtual-appliance create"] = VirtualApplianceCreate(loader=self)
    self.command_table["network virtual-appliance update"] = VirtualApplianceUpdate(loader=self)
    # endregion

    # region CustomIp
    from .custom import CustomIpPrefixCreate, CustomIpPrefixUpdate
    self.command_table["network custom-ip prefix create"] = CustomIpPrefixCreate(loader=self)
    self.command_table["network custom-ip prefix update"] = CustomIpPrefixUpdate(loader=self)
    # endregion

    # region DdosCustomPolicy
    with self.command_group('network ddos-custom-policy') as g:
        g.custom_command('create', 'create_ddos_custom_policy', supports_no_wait=True)
    # endregion



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from time import sleep

from azure.cli.core.auth.util import decode_access_token
from azure.cli.core.azclierror import AuthenticationError
from azure.cli.testsdk import LiveScenarioTest

ARM_URL = "https://eastus2euap.management.azure.com/"  # ARM canary
ARM_MAX_RETRY = 30
ARM_RETRY_INTERVAL = 10


class ConditionalAccessScenarioTest(LiveScenarioTest):

    def setUp(self):
        super().setUp()
        # Clear MSAL cache to avoid unexpected tokens from cache
        self.cmd('az account clear')

    def test_conditional_access_mfa(self):
        """
        This test should be run using a user account that
          - doesn't require MFA for ARM
          - requires MFA for data-plane resource

        The result ATs are checked per
          Microsoft identity platform access tokens
          https://learn.microsoft.com/en-us/azure/active-directory/develop/access-tokens

        Following claims are checked:
          - aud (Audience): https://tools.ietf.org/html/rfc7519#section-4.1.3
          - amr (Authentication Method Reference): https://tools.ietf.org/html/rfc8176
        """

        scope = 'https://pas.windows.net/CheckMyAccess/Linux/.default'
        self.kwargs['scope'] = scope

        # region non-MFA session

        # Login to ARM (MFA not required)
        # In the browser, if the user already exists, make sure to logout first and re-login to clear browser cache
        self.cmd('az login')

        # Getting ARM AT and check claims
        result = self.cmd('az account get-access-token').get_output_in_json()
        decoded = decode_access_token(result['accessToken'])
        assert decoded['aud'] == self.cli_ctx.cloud.endpoints.active_directory_resource_id
        assert decoded['amr'] == ['pwd']

        # Getting data-plane AT with ARM RT (step-up) fails
        with self.assertRaises(AuthenticationError) as cm:
            self.cmd('az account get-access-token --scope {scope}')

        # Check re-login recommendation
        re_login_command = 'az login --scope {scope}'.format(**self.kwargs)
        assert 'AADSTS50076' in cm.exception.error_msg
        assert re_login_command in cm.exception.recommendations[0]

        # endregion

        # region MFA session

        # Re-login with data-plane scope (MFA required)
        # Getting ARM AT with data-plane RT (step-down) succeeds
        self.cmd(re_login_command)

        # Getting ARM AT and check claims
        result = self.cmd('az account get-access-token').get_output_in_json()
        decoded = decode_access_token(result['accessToken'])
        assert decoded['aud'] == self.cli_ctx.cloud.endpoints.active_directory_resource_id
        assert decoded['amr'] == ['pwd']

        # Getting data-plane AT and check claims
        result = self.cmd('az account get-access-token --scope {scope}').get_output_in_json()
        decoded = decode_access_token(result['accessToken'])
        assert decoded['aud'] in scope
        assert decoded['amr'] == ['pwd', 'mfa']

        self.cmd('logout')
        # endregion


class CAEScenarioTest(LiveScenarioTest):

    def setUp(self):
        super().setUp()
        # Clear MSAL cache to avoid unexpected tokens from cache
        self.cmd('az account clear')

    def tearDown(self):
        self.cmd('az account clear')

    def _retry_until_error(self, cmd):
        remaining_reties = ARM_MAX_RETRY
        while remaining_reties > 0:
            remaining_reties -= 1
            sleep(ARM_RETRY_INTERVAL)
            self.cmd(cmd)
        raise AssertionError("Retry chance exhausted.")

    def test_client_capabilities(self):
        # Verify the access token has CAE enabled
        result = self.cmd('account get-access-token').get_output_in_json()
        access_token = result['accessToken']
        decoded = decode_access_token(access_token)
        self.assertEqual(decoded['xms_cc'], ['CP1'])  # xms_cc: extension microsoft client capabilities

    def test_cae_scenario(self):
        arm_command = "group list"

        self.cmd('login')
        self.test_client_capabilities()

        # Test access token is working
        self.cmd(arm_command)

        self._revoke_sign_in_sessions()

        # Keep trying until failure
        with self.assertRaises(AuthenticationError) as cm:
            self._retry_until_error(arm_command)

        assert 'AADSTS50173' in cm.exception.error_msg
        assert 'az login' in cm.exception.recommendations[0]

    def _revoke_sign_in_sessions(self):
        # Manually revoke sign-in sessions
        self.cmd('rest -m POST -u https://graph.microsoft.com/v1.0/me/revokeSignInSessions')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import json
import os
import re
import sys

from azure.cli.core._environment import get_config_dir
from knack.log import get_logger
from knack.util import CLIError
from msal import PublicClientApplication, ConfidentialClientApplication

from .constants import AZURE_CLI_CLIENT_ID
from .msal_credentials import UserCredential, ServicePrincipalCredential
from .persistence import load_persisted_token_cache, file_extensions, load_secret_store
from .util import check_result

# Service principal entry properties. Names are taken from OAuth 2.0 client credentials flow parameters:
# https://learn.microsoft.com/en-us/entra/identity-platform/v2-oauth2-client-creds-grant-flow
_TENANT = 'tenant'
_CLIENT_ID = 'client_id'
_CLIENT_SECRET = 'client_secret'
_CERTIFICATE = 'certificate'
_USE_CERT_SN_ISSUER = 'use_cert_sn_issuer'
_CLIENT_ASSERTION = 'client_assertion'

# For environment credential
AZURE_AUTHORITY_HOST = "AZURE_AUTHORITY_HOST"
AZURE_TENANT_ID = "AZURE_TENANT_ID"
AZURE_CLIENT_ID = "AZURE_CLIENT_ID"
AZURE_CLIENT_SECRET = "AZURE_CLIENT_SECRET"

WAM_PROMPT = (
    "Select the account you want to log in with. "
    "For more information on login with Azure CLI, see https://go.microsoft.com/fwlink/?linkid=2271136")

logger = get_logger(__name__)


class Identity:  # pylint: disable=too-many-instance-attributes
    """Manage user or service principal identities and log into Microsoft identity platform.
    """

    # MSAL token cache.
    # It follows singleton pattern so that all MSAL app instances share the same token cache.
    _msal_token_cache = None

    # MSAL HTTP cache for MSAL's tenant discovery, retry-after error cache, etc.
    # It *must* follow singleton pattern so that all MSAL app instances share the same HTTP cache.
    # https://github.com/AzureAD/microsoft-authentication-library-for-python/pull/407
    _msal_http_cache = None

    # Instance of ServicePrincipalStore.
    # It follows singleton pattern so that _secret_file is read only once.
    _service_principal_store_instance = None

    def __init__(self, authority, tenant_id=None, client_id=None, encrypt=False, use_msal_http_cache=True,
                 enable_broker_on_windows=None, instance_discovery=None):
        """
        :param authority: Authentication authority endpoint. For example,
            - AAD: https://login.microsoftonline.com
            - ADFS: https://adfs.redmond.azurestack.corp.microsoft.com/adfs
        :param tenant_id: Tenant GUID, like 00000000-0000-0000-0000-000000000000. If unspecified, default to
            'organizations'.
        :param client_id: Client ID of the CLI application.
        :param encrypt:  Whether to encrypt MSAL token cache and service principal entries.
        """
        self.authority = authority
        self.tenant_id = tenant_id
        self.client_id = client_id or AZURE_CLI_CLIENT_ID
        self._encrypt = encrypt
        self._use_msal_http_cache = use_msal_http_cache
        self._enable_broker_on_windows = enable_broker_on_windows
        self._instance_discovery = instance_discovery

        # Build the authority in MSAL style
        self._msal_authority, self._is_adfs = _get_authority_url(authority, tenant_id)

        config_dir = get_config_dir()
        self._token_cache_file = os.path.join(config_dir, "msal_token_cache")
        self._secret_file = os.path.join(config_dir, "service_principal_entries")
        self._msal_http_cache_file = os.path.join(config_dir, "msal_http_cache.bin")

        # We make _msal_app_instance an instance attribute, instead of a class attribute,
        # because MSAL apps can have different tenant IDs.
        self._msal_app_instance = None

    @property
    def _msal_app_kwargs(self):
        """kwargs for creating ClientApplication (including its subclass ConfidentialClientApplication).
        MSAL token cache and HTTP cache are lazily created.
        """
        if not Identity._msal_token_cache:
            Identity._msal_token_cache = self._load_msal_token_cache()

        if self._use_msal_http_cache and not Identity._msal_http_cache:
            Identity._msal_http_cache = self._load_msal_http_cache()

        return {
            "authority": self._msal_authority,
            "token_cache": Identity._msal_token_cache,
            "http_cache": Identity._msal_http_cache,
            "instance_discovery": self._instance_discovery,
            # CP1 means we can handle claims challenges (CAE)
            "client_capabilities": None if "AZURE_IDENTITY_DISABLE_CP1" in os.environ else ["CP1"]
        }

    @property
    def _msal_public_app_kwargs(self):
        """kwargs for creating PublicClientApplication."""
        # enable_broker_on_windows can only be used on PublicClientApplication.
        return {**self._msal_app_kwargs,
                "enable_broker_on_windows": self._enable_broker_on_windows,
                "enable_pii_log": True}

    @property
    def _msal_app(self):
        """A PublicClientApplication instance for user login/logout.
        The instance is lazily created.
        """
        if not self._msal_app_instance:
            self._msal_app_instance = PublicClientApplication(self.client_id, **self._msal_public_app_kwargs)
        return self._msal_app_instance

    def _load_msal_token_cache(self):
        # Store for user token persistence
        cache = load_persisted_token_cache(self._token_cache_file, self._encrypt)
        return cache

    def _load_msal_http_cache(self):
        from .binary_cache import BinaryCache
        http_cache = BinaryCache(self._msal_http_cache_file)
        return http_cache

    @property
    def _service_principal_store(self):
        """A ServicePrincipalStore instance for service principal entries persistence.
        The instance is lazily created.
        """
        if not Identity._service_principal_store_instance:
            store = load_secret_store(self._secret_file, self._encrypt)
            Identity._service_principal_store_instance = ServicePrincipalStore(store)
        return Identity._service_principal_store_instance

    def login_with_auth_code(self, scopes, claims_challenge=None):
        # Emit a warning to inform that a browser is opened.
        # Only show the path part of the URL and hide the query string.

        def _prompt_launching_ui(ui=None, **_):
            if ui == 'browser':
                logger.warning("A web browser has been opened at %s. Please continue the login in the web browser. "
                               "If no web browser is available or if the web browser fails to open, use device code "
                               "flow with `az login --use-device-code`.",
                               self._msal_app.authority.authorization_endpoint)
            elif ui == 'broker':
                logger.warning(WAM_PROMPT)

        from .util import read_response_templates
        success_template, error_template = read_response_templates()

        # For AAD, use port 0 to let the system choose arbitrary unused ephemeral port to avoid port collision
        # on port 8400 from the old design. However, ADFS only allows port 8400.
        result = self._msal_app.acquire_token_interactive(
            scopes, prompt='select_account', port=8400 if self._is_adfs else None,
            success_template=success_template, error_template=error_template,
            parent_window_handle=self._msal_app.CONSOLE_WINDOW_HANDLE, on_before_launching_ui=_prompt_launching_ui,
            enable_msa_passthrough=True,
            claims_challenge=claims_challenge)
        return check_result(result)

    def login_with_device_code(self, scopes, claims_challenge=None):
        flow = self._msal_app.initiate_device_flow(scopes, claims_challenge=claims_challenge)
        if "user_code" not in flow:
            raise ValueError(
                "Fail to create device flow. Err: %s" % json.dumps(flow, indent=4))
        from azure.cli.core.style import print_styled_text, Style
        print_styled_text((Style.WARNING, flow["message"]), file=sys.stderr)
        result = self._msal_app.acquire_token_by_device_flow(flow, claims_challenge=claims_challenge)
        return check_result(result)

    def login_with_username_password(self, username, password, scopes):
        result = self._msal_app.acquire_token_by_username_password(username, password, scopes)
        return check_result(result)

    def login_with_service_principal(self, client_id, credential, scopes):
        """
        `credential` is a dict returned by ServicePrincipalAuth.build_credential
        """
        sp_auth = ServicePrincipalAuth.build_from_credential(self.tenant_id, client_id, credential)
        client_credential = sp_auth.get_msal_client_credential()
        cred = ServicePrincipalCredential(client_id, client_credential, **self._msal_app_kwargs)
        cred.acquire_token(scopes)

        # Only persist the service principal after a successful login
        entry = sp_auth.get_entry_to_persist()
        self._service_principal_store.save_entry(entry)

    def logout_user(self, username):
        # If username is an SP client ID, it is ignored
        accounts = self._msal_app.get_accounts(username)
        for account in accounts:
            self._msal_app.remove_account(account)

    def logout_all_users(self):
        # Remove users from MSAL
        accounts = self._msal_app.get_accounts()
        for account in accounts:
            self._msal_app.remove_account(account)

        # Also remove token cache file
        for e in file_extensions.values():
            _try_remove(self._token_cache_file + e)

    def logout_service_principal(self, client_id):
        # If client_id is a username, it is ignored

        # Step 1: Remove SP from MSAL token cache
        # Note that removing SP access tokens shouldn't rely on SP store
        cca = ConfidentialClientApplication(client_id, **self._msal_app_kwargs)
        cca.remove_tokens_for_client()

        # Step 2: Remove SP from SP store
        self._service_principal_store.remove_entry(client_id)

    def logout_all_service_principal(self):
        # remove service principal secrets
        # TODO: As MSAL provides no interface to get all service principals in its token cache, this method can't
        #   clear all service principals' access tokens from MSAL token cache.
        for e in file_extensions.values():
            _try_remove(self._secret_file + e)

    def get_user(self, user=None):
        accounts = self._msal_app.get_accounts(user) if user else self._msal_app.get_accounts()
        return accounts

    def get_user_credential(self, username):
        return UserCredential(self.client_id, username, **self._msal_public_app_kwargs)

    def get_service_principal_credential(self, client_id):
        entry = self._service_principal_store.load_entry(client_id, self.tenant_id)
        client_credential = ServicePrincipalAuth(entry).get_msal_client_credential()
        return ServicePrincipalCredential(client_id, client_credential, **self._msal_app_kwargs)


class ServicePrincipalAuth:  # pylint: disable=too-many-instance-attributes
    def __init__(self, entry):
        # Initialize all attributes first, so that we don't need to call getattr to check their existence
        self.client_id = None
        self.tenant = None
        # secret
        self.client_secret = None
        # certificate
        self.certificate = None
        self.use_cert_sn_issuer = None
        # federated identity credential
        self.client_assertion = None

        # Internal attributes for certificate
        # They are computed at runtime and not persisted in the service principal entry.
        self._certificate_string = None
        self._thumbprint = None
        self._public_certificate = None

        self.__dict__.update(entry)

        if self.certificate:
            from OpenSSL.crypto import load_certificate, FILETYPE_PEM, Error
            try:
                with open(self.certificate, 'r') as file_reader:
                    self._certificate_string = file_reader.read()
                    cert = load_certificate(FILETYPE_PEM, self._certificate_string)
                    self._thumbprint = cert.digest("sha1").decode().replace(':', '')
                    if entry.get(_USE_CERT_SN_ISSUER):
                        # low-tech but safe parsing based on
                        # https://github.com/libressl-portable/openbsd/blob/master/src/lib/libcrypto/pem/pem.h
                        match = re.search(r'-----BEGIN CERTIFICATE-----(?P<cert_value>[^-]+)-----END CERTIFICATE-----',
                                          self._certificate_string, re.I)
                        self._public_certificate = match.group()
            except (UnicodeDecodeError, Error) as ex:
                raise CLIError('Invalid certificate, please use a valid PEM file. Error detail: {}'.format(ex))

    @classmethod
    def build_from_credential(cls, tenant_id, client_id, credential):
        entry = {
            _TENANT: tenant_id,
            _CLIENT_ID: client_id
        }
        entry.update(credential)
        return ServicePrincipalAuth(entry)

    @classmethod
    def build_credential(cls, client_secret=None,
                         certificate=None, use_cert_sn_issuer=None,
                         client_assertion=None):
        """Build credential from user input. The credential looks like below, but only one key can exist.
        {
            'client_secret': 'my_secret',
            'certificate': '/path/to/cert.pem',
            'client_assertion': 'my_federated_token'
        }
        """
        entry = {}
        if client_secret:
            entry[_CLIENT_SECRET] = client_secret
        elif certificate:
            entry[_CERTIFICATE] = os.path.expanduser(certificate)
            if use_cert_sn_issuer:
                entry[_USE_CERT_SN_ISSUER] = use_cert_sn_issuer
        elif client_assertion:
            entry[_CLIENT_ASSERTION] = client_assertion
        return entry

    def get_entry_to_persist(self):
        """Get a service principal entry that can be persisted by ServicePrincipalStore."""
        persisted_keys = [_CLIENT_ID, _TENANT, _CLIENT_SECRET, _CERTIFICATE, _USE_CERT_SN_ISSUER, _CLIENT_ASSERTION]
        # Only persist certain attributes whose values are not None
        return {k: v for k, v in self.__dict__.items() if k in persisted_keys and v}

    def get_msal_client_credential(self):
        """Get a client_credential that can be consumed by msal.ConfidentialClientApplication."""
        client_credential = None

        # client_secret
        # "your client secret"
        if self.client_secret:
            client_credential = self.client_secret

        # certificate
        # {
        #     "private_key": "...-----BEGIN PRIVATE KEY-----... in PEM format",
        #     "thumbprint": "A1B2C3D4E5F6...",
        #     "public_certificate": "...-----BEGIN CERTIFICATE-----...",
        # }
        if self.certificate:
            client_credential = {
                "private_key": self._certificate_string,
                "thumbprint": self._thumbprint
            }
            if self._public_certificate:
                client_credential['public_certificate'] = self._public_certificate

        # client_assertion
        # {
        #     "client_assertion": "...a JWT with claims aud, exp, iss, jti, nbf, and sub..."
        # }
        if self.client_assertion:
            client_credential = {'client_assertion': self.client_assertion}

        return client_credential


class ServicePrincipalStore:
    """Save secrets in MSAL custom secret store for Service Principal authentication.
    """

    def __init__(self, secret_store):
        self._secret_store = secret_store
        self._entries = []

    def load_entry(self, sp_id, tenant):
        self._load_persistence()
        matched = [x for x in self._entries if sp_id == x[_CLIENT_ID]]
        if not matched:
            raise CLIError("Could not retrieve credential from local cache for service principal {}. "
                           "Run `az login` for this service principal."
                           .format(sp_id))
        matched_with_tenant = [x for x in matched if tenant == x[_TENANT]]
        if matched_with_tenant:
            cred = matched_with_tenant[0]
        else:
            logger.warning("Could not retrieve credential from local cache for service principal %s under tenant %s. "
                           "Trying credential under tenant %s, assuming that is an app credential.",
                           sp_id, tenant, matched[0][_TENANT])
            cred = matched[0]

        return cred

    def save_entry(self, sp_entry):
        self._load_persistence()

        self._entries = [
            x for x in self._entries
            if not (sp_entry[_CLIENT_ID] == x[_CLIENT_ID] and
                    sp_entry[_TENANT] == x[_TENANT])]

        self._entries.append(sp_entry)
        self._save_persistence()

    def remove_entry(self, sp_id):
        self._load_persistence()
        state_changed = False

        # clear service principal creds
        matched = [x for x in self._entries
                   if x[_CLIENT_ID] == sp_id]
        if matched:
            state_changed = True
            self._entries = [x for x in self._entries
                             if x not in matched]

        if state_changed:
            self._save_persistence()

    def _save_persistence(self):
        self._secret_store.save(self._entries)

    def _load_persistence(self):
        self._entries = self._secret_store.load()


def _get_authority_url(authority_endpoint, tenant):
    """Convert authority endpoint (active_directory) to MSAL authority:
        - AAD: https://login.microsoftonline.com/your_tenant
        - ADFS: https://adfs.redmond.azurestack.corp.microsoft.com/adfs
    For ADFS, tenant is discarded.
    """

    # Some Azure Stack (bellevue)'s metadata returns
    #   "loginEndpoint": "https://login.microsoftonline.com/"
    # Normalize it by removing the trailing /, so that authority_url won't become
    # "https://login.microsoftonline.com//tenant_id".
    authority_endpoint = authority_endpoint.rstrip('/').lower()
    is_adfs = authority_endpoint.endswith('adfs')
    if is_adfs:
        authority_url = authority_endpoint
    else:
        authority_url = '{}/{}'.format(authority_endpoint, tenant or "organizations")
    return authority_url, is_adfs


def _try_remove(path):
    try:
        os.remove(path)
    except FileNotFoundError:
        pass


def get_environment_credential():
    # A temporary workaround used by rdbms module to use environment credential.
    # TODO: Integrate with Identity and utilize MSAL HTTP and token cache to officially implement
    #  https://github.com/Azure/azure-cli/issues/10241
    from os import getenv

    sp_auth = ServicePrincipalAuth({
        _TENANT: getenv(AZURE_TENANT_ID),
        _CLIENT_ID: getenv(AZURE_CLIENT_ID),
        _CLIENT_SECRET: getenv(AZURE_CLIENT_SECRET)
    })

    authority, _ = _get_authority_url(
        # Override authority host if defined as env var
        getenv(AZURE_AUTHORITY_HOST) or 'https://login.microsoftonline.com',
        getenv(AZURE_TENANT_ID))
    credentials = ServicePrincipalCredential(sp_auth, authority=authority)
    return credentials



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import unittest
from unittest import mock
from azure.cli.core.aaz._command_ctx import AAZCommandCtx
from azure.cli.core.aaz import AAZArgumentsSchema
from azure.cli.core.mock import DummyCli
import tempfile

_METADATA = [
    {
        "name": "TestCloud",
        "portal": "https://portal.azure.com.test",
        "authentication": {
            "loginEndpoint": "https://login.microsoftonline.com.test/",
            "audiences": [
                "https://management.core.windows.net.test/",
                "https://management.azure.com.test/"
            ],
            "tenant": "common.test",
            "identityProvider": "AAD.test"
        },
        "media": "https://rest.media.azure.net.test",
        "graphAudience": "https://graph.windows.net.test/",
        "graph": "https://graph.windows.net.test/",
        "suffixes": {
            "azureDataLakeStoreFileSystem": "azuredatalakestore.net.test",
            "acrLoginServer": "azurecr.io.test",
            "sqlServerHostname": "database.windows.net.test",
            "azureDataLakeAnalyticsCatalogAndJob": "azuredatalakeanalytics.net.test",
            "keyVaultDns": "vault.azure.net.test",
            "storage": "core.windows.net.test",
            "azureFrontDoorEndpointSuffix": "azurefd.net.test",
            "notExist": "notExistSuffix.test",
        },
        "batch": "https://batch.core.windows.net.test/",
        "resourceManager": "https://management.azure.com.test/",
        "vmImageAliasDoc": "https://raw.githubusercontent.com.test/Azure/azure-rest-api-specs/master/arm-compute/quickstart-templates/aliases.json",
        "activeDirectoryDataLake": "https://datalake.azure.net.test/",
        "sqlManagement": "https://management.core.windows.net.test:8443/",
        "gallery": "https://gallery.azure.com.test/",
        "notExist": {
            "Special": "https://notexist.azure.com.test/",
        }
    }
]


class TestAAZCommandCtx(unittest.TestCase):

    @mock.patch('azure.cli.core.cloud.retrieve_arm_cloud_metadata', return_value=_METADATA)
    @mock.patch('azure.cli.core.cloud.CLOUD_CONFIG_FILE', tempfile.mkstemp()[1])
    @mock.patch('azure.cli.core.cloud.GLOBAL_CONFIG_DIR', tempfile.mkdtemp())
    @mock.patch.dict('os.environ', {'ARM_CLOUD_METADATA_URL': 'https://management.azure.com/metadata/endpoints?api-version=2019-05-01'})
    def test_retrieve_value_in_arm_cloud_metadata(self, *args, **kwargs):
        from azure.cli.core.aaz import AAZBaseClient
        from azure.cli.core.cloud import get_active_cloud_name, refresh_known_clouds, _set_active_cloud, CloudSuffixes, CloudEndpoints
        from azure.cli.core.aaz import AAZStrArg
        from azure.cli.core.aaz import exceptions
        schema = AAZArgumentsSchema()
        schema.region = AAZStrArg()
        region = "global"
        ctx = AAZCommandCtx(
            cli_ctx=DummyCli(random_config_dir=True),
            schema=schema,
            command_args={
                "region": region,
            }
        )
        refresh_known_clouds()
        self.assertEqual(get_active_cloud_name(ctx.cli_ctx), 'AzureCloud')

        self.assertEqual(
            AAZBaseClient._retrieve_value_in_arm_cloud_metadata(ctx, "notExist.Special"),
            None)

        _set_active_cloud(ctx.cli_ctx, 'TestCloud')

        self.assertEqual(get_active_cloud_name(ctx.cli_ctx), 'TestCloud')
        self.assertEqual(
            AAZBaseClient._retrieve_value_in_arm_cloud_metadata(ctx, "suffixes.azureDataLakeStoreFileSystem"),
            "azuredatalakestore.net.test")
        self.assertEqual(
            AAZBaseClient._retrieve_value_in_arm_cloud_metadata(ctx, "notExist.Special"),
            "https://notexist.azure.com.test/")
        self.assertEqual(
            AAZBaseClient._retrieve_value_in_arm_cloud_metadata(ctx, "authentication.audiences[0]"),
            "https://management.core.windows.net.test/")
        self.assertEqual(
            AAZBaseClient._retrieve_value_in_arm_cloud_metadata(ctx, "authentication.audiences[100]"),
            None)
        self.assertEqual(
            AAZBaseClient._retrieve_value_in_arm_cloud_metadata(ctx, "authentication.audiences100"),
            None)

        with self.assertRaises(TypeError):
            AAZBaseClient._retrieve_value_in_arm_cloud_metadata(ctx, "authentication.audiences.1")

        with self.assertRaises(exceptions.AAZInvalidShorthandSyntaxError):
            AAZBaseClient._retrieve_value_in_arm_cloud_metadata(ctx, "authentication.audiences[0")

        for arm_idx in CloudSuffixes.ARM_METADATA_INDEX.values():
            value = AAZBaseClient.get_cloud_suffix(ctx, arm_idx)
            self.assertTrue(value is None or value.startswith("."))

        self.assertEqual(AAZBaseClient.get_cloud_suffix(ctx, "suffixes.notExist"),
                         ".notExistSuffix.test")

        for arm_idx in CloudEndpoints.ARM_METADATA_INDEX.values():
            value = AAZBaseClient.get_cloud_endpoint(ctx, arm_idx)
            self.assertTrue(value is None or value.startswith("https://"))

        self.assertEqual(AAZBaseClient.get_cloud_endpoint(ctx, "notExist.Special"),
                         "https://notexist.azure.com.test/")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import json
import os

from azure.cli.testsdk.scenario_tests import RecordingProcessor
from azure.cli.testsdk.scenario_tests.utilities import is_json_payload
from azure.cli.core.util import shell_safe_json_parse

def create_config_store(test, kwargs):
    if 'retention_days' not in kwargs:
        kwargs.update({
            'retention_days': 1
        })
    test.cmd('appconfig create -n {config_store_name} -g {rg} -l {rg_loc} --sku {sku} --retention-days {retention_days}')


def _get_local_test_resource_prefix():
    return os.environ.get("AZURE_CLI_LOCAL_TEST_RESOURCE_PREFIX")

def get_resource_name_prefix(prefix):
    resource_prefix = _get_local_test_resource_prefix()
    return prefix if resource_prefix is None else resource_prefix + prefix

class CredentialResponseSanitizer(RecordingProcessor):
    def process_response(self, response):
        if is_json_payload(response):
            try:
                json_data = shell_safe_json_parse(response["body"]["string"])

                if isinstance(json_data.get("value"), list):
                    for idx, credential in enumerate(json_data["value"]):
                        self._try_replace_secret(credential, idx)

                    response["body"]["string"] = json.dumps(json_data)
                
                elif isinstance(json_data, dict):
                    self._try_replace_secret(json_data)

                    response["body"]["string"] = json.dumps(json_data)

            except Exception:
                pass

        return response

    def _try_replace_secret(self, credential, idx = 0):
        if "connectionString" in credential:
            credential["id"] = "sanitized_id{}".format(idx + 1)
            credential["value"] = "sanitized_secret{}".format(idx + 1)

            endpoint = next(
                filter(lambda x: x.startswith("Endpoint="), credential["connectionString"].split(";")))[len("Endpoint="):]

            credential["connectionString"] = "Endpoint={};Id={};Secret={}".format(
                endpoint, credential["id"], credential["value"])



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=broad-exception-caught, line-too-long, no-else-return
from typing import Any, Dict

from azure.cli.core.commands import AzCliCommand
from knack.util import CLIError

from ._client_factory import handle_raw_exception
from ._utils import register_provider_if_needed, _validate_subscription_registered


class BaseResource:
    def __init__(
        self, cmd: AzCliCommand, client: Any, raw_parameters: Dict, models: str
    ):
        self.raw_param = raw_parameters
        self.cmd = cmd
        self.client = client
        self.models = models

    def register_provider(self, *rp_name_list):
        for rp in rp_name_list:
            register_provider_if_needed(self.cmd, rp)

    def validate_subscription_registered(self, *rp_name_list):
        for rp in rp_name_list:
            _validate_subscription_registered(self.cmd, rp)

    def list(self):
        try:
            if self.get_argument_resource_group_name() is None:
                return self.client.list_by_subscription(cmd=self.cmd)
            else:
                return self.client.list_by_resource_group(cmd=self.cmd, resource_group_name=self.get_argument_resource_group_name())
        except CLIError as e:
            handle_raw_exception(e)

    def show(self):
        try:
            return self.client.show(cmd=self.cmd, resource_group_name=self.get_argument_resource_group_name(), name=self.get_argument_name())
        except CLIError as e:
            handle_raw_exception(e)

    def delete(self):
        try:
            return self.client.delete(cmd=self.cmd, name=self.get_argument_name(), resource_group_name=self.get_argument_resource_group_name(), no_wait=self.get_argument_no_wait())
        except CLIError as e:
            handle_raw_exception(e)

    def get_param(self, key) -> Any:
        return self.raw_param.get(key)

    def set_param(self, key, value):
        self.raw_param[key] = value

    def get_argument_name(self):
        return self.get_param("name")

    def get_argument_resource_group_name(self):
        return self.get_param("resource_group_name")

    def get_argument_no_wait(self):
        return self.get_param("no_wait")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.testsdk import ScenarioTest, ResourceGroupPreparer


class AcrNetworkRuleCommandsTests(ScenarioTest):

    @ResourceGroupPreparer(location='westcentralus')
    def test_acr_network_rule(self, resource_group, resource_group_location):
        self.kwargs.update({
            'acr_provider': 'Microsoft.ContainerRegistry',
            'registry_name': self.create_random_name('clireg', 20),
            'rg_loc': resource_group_location,
            'sku': 'Premium',
            'allow_action': 'Allow',
            'deny_action': 'Deny',
            'vnet_name': 'cliregvnet',
            'subnet': 'cliregsubnet',
            'ip_address': '16.17.18.0/24'
        })

        # Create vnet and subnet
        self.cmd('network vnet create -g {rg} -n {vnet_name} --subnet-name {subnet}',
                 checks=[self.check('newVNet.name', '{vnet_name}'),
                         self.check('newVNet.subnets[0].name', '{subnet}')])

        # Enable service endpoints
        response = self.cmd('az network vnet subnet update -g {rg} --vnet-name {vnet_name} -n {subnet} --service-endpoints {acr_provider}',
                            checks=[self.check('name', '{subnet}'),
                                    self.check('serviceEndpoints[0].service', '{acr_provider}')]).get_output_in_json()

        subnet_id = response['id']

        # Create a registry
        self.cmd('acr create -n {registry_name} -g {rg} -l {rg_loc} --sku {sku} --default-action {deny_action}',
                 checks=[self.check('name', '{registry_name}'),
                         self.check('location', '{rg_loc}'),
                         self.check('adminUserEnabled', False),
                         self.check('sku.name', '{sku}'),
                         self.check('sku.tier', '{sku}'),
                         self.check('provisioningState', 'Succeeded'),
                         self.check('networkRuleSet.defaultAction', '{deny_action}'),
                         self.check('networkRuleSet.ipRules', [])])

        # Add a VNet rule
        self.cmd('acr network-rule add -g {rg} -n {registry_name} --vnet-name {vnet_name} --subnet {subnet}',
                 checks=[self.check('name', '{registry_name}'),
                         self.check('provisioningState', 'Succeeded'),
                         self.check('networkRuleSet.defaultAction', '{deny_action}'),
                         self.check('networkRuleSet.virtualNetworkRules[0].virtualNetworkResourceId', subnet_id),
                         self.check('networkRuleSet.virtualNetworkRules[0].action', '{allow_action}'),
                         self.check('networkRuleSet.ipRules', [])])

        # Add an IP rule
        self.cmd('acr network-rule add -g {rg} -n {registry_name} --ip-address {ip_address}',
                 checks=[self.check('name', '{registry_name}'),
                         self.check('provisioningState', 'Succeeded'),
                         self.check('networkRuleSet.defaultAction', '{deny_action}'),
                         self.check('networkRuleSet.virtualNetworkRules[0].virtualNetworkResourceId', subnet_id),
                         self.check('networkRuleSet.virtualNetworkRules[0].action', '{allow_action}'),
                         self.check('networkRuleSet.ipRules[0].ipAddressOrRange', '{ip_address}'),
                         self.check('networkRuleSet.ipRules[0].action', '{allow_action}')])

        self.cmd('acr network-rule list -g {rg} -n {registry_name}',
                 checks=[self.check('virtualNetworkRules[0].virtualNetworkResourceId', subnet_id),
                         self.check('virtualNetworkRules[0].action', '{allow_action}'),
                         self.check('ipRules[0].ipAddressOrRange', '{ip_address}'),
                         self.check('ipRules[0].action', '{allow_action}')])

        # Switch network rule default action
        self.cmd('acr update -n {registry_name} -g {rg} --default-action {allow_action}',
                 checks=[self.check('name', '{registry_name}'),
                         self.check('provisioningState', 'Succeeded'),
                         self.check('networkRuleSet.defaultAction', '{allow_action}'),
                         self.check('networkRuleSet.ipRules[0].ipAddressOrRange', '{ip_address}'),
                         self.check('networkRuleSet.ipRules[0].action', '{allow_action}')])

        # Remove a VNet rule
        self.cmd('acr network-rule remove -g {rg} -n {registry_name} --vnet-name {vnet_name} --subnet {subnet}',
                 checks=[self.check('name', '{registry_name}'),
                         self.check('provisioningState', 'Succeeded'),
                         self.check('networkRuleSet.defaultAction', '{allow_action}'),
                         self.check('networkRuleSet.virtualNetworkRules', []),
                         self.check('networkRuleSet.ipRules[0].ipAddressOrRange', '{ip_address}'),
                         self.check('networkRuleSet.ipRules[0].action', '{allow_action}')])

        # Remove an IP rule
        self.cmd('acr network-rule remove -g {rg} -n {registry_name} --ip-address {ip_address}',
                 checks=[self.check('name', '{registry_name}'),
                         self.check('provisioningState', 'Succeeded'),
                         self.check('networkRuleSet.defaultAction', '{allow_action}'),
                         self.check('networkRuleSet.virtualNetworkRules', []),
                         self.check('networkRuleSet.ipRules', [])])

        self.cmd('acr network-rule list -g {rg} -n {registry_name}',
                 checks=[self.check('virtualNetworkRules', []),
                         self.check('ipRules', [])])

        self.cmd('acr delete -g {rg} -n {registry_name} -y')
        self.cmd('network vnet delete -g {rg} -n {vnet_name}')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=unused-argument, line-too-long
from azure.cli.core.util import user_confirmation
from datetime import datetime
from knack.log import get_logger
from knack.util import CLIError
from ..utils.validators import validate_public_access_server, validate_resource_group

logger = get_logger(__name__)
# pylint: disable=raise-missing-from


def firewall_rule_delete_func(cmd, client, resource_group_name, server_name, firewall_rule_name, yes=None):
    validate_resource_group(resource_group_name)
    validate_public_access_server(cmd, resource_group_name, server_name)

    result = None
    if not yes:
        user_confirmation(
            "Are you sure you want to delete the firewall-rule '{0}' in server '{1}', resource group '{2}'".format(
                firewall_rule_name, server_name, resource_group_name))
    try:
        result = client.begin_delete(resource_group_name, server_name, firewall_rule_name)
    except Exception as ex:  # pylint: disable=broad-except
        logger.error(ex)
    return result


def firewall_rule_create_func(cmd, client, resource_group_name, server_name, firewall_rule_name=None, start_ip_address=None, end_ip_address=None):
    validate_resource_group(resource_group_name)
    validate_public_access_server(cmd, resource_group_name, server_name)

    if end_ip_address is None and start_ip_address is not None:
        end_ip_address = start_ip_address
    elif start_ip_address is None and end_ip_address is not None:
        start_ip_address = end_ip_address
    elif start_ip_address is None and end_ip_address is None:
        raise CLIError("Incorrect Usage : Need to pass in value for either \'--start-ip-address\' or \'--end-ip-address\'.")

    if firewall_rule_name is None:
        now = datetime.now()
        firewall_rule_name = 'FirewallIPAddress_{}-{}-{}_{}-{}-{}'.format(now.year, now.month, now.day, now.hour, now.minute,
                                                                          now.second)
        if start_ip_address == '0.0.0.0' and end_ip_address == '0.0.0.0':
            logger.warning('Configuring server firewall rule, \'azure-access\', to accept connections from all '
                           'Azure resources...')
            firewall_rule_name = 'AllowAllAzureServicesAndResourcesWithinAzureIps_{}-{}-{}_{}-{}-{}'.format(now.year, now.month,
                                                                                                            now.day, now.hour,
                                                                                                            now.minute, now.second)
        elif start_ip_address == end_ip_address:
            logger.warning('Configuring server firewall rule to accept connections from \'%s\'...', start_ip_address)
        else:
            if start_ip_address == '0.0.0.0' and end_ip_address == '255.255.255.255':
                firewall_rule_name = 'AllowAll_{}-{}-{}_{}-{}-{}'.format(now.year, now.month, now.day,
                                                                         now.hour, now.minute, now.second)
            logger.warning('Configuring server firewall rule to accept connections from \'%s\' to \'%s\'...', start_ip_address,
                           end_ip_address)

    parameters = {
        'properties': {
            'startIpAddress': start_ip_address,
            'endIpAddress': end_ip_address
        }
    }

    return client.begin_create_or_update(
        resource_group_name,
        server_name,
        firewall_rule_name,
        parameters)


def flexible_firewall_rule_custom_getter(cmd, client, resource_group_name, server_name, firewall_rule_name):
    validate_resource_group(resource_group_name)
    validate_public_access_server(cmd, resource_group_name, server_name)
    return client.get(resource_group_name, server_name, firewall_rule_name)


def flexible_firewall_rule_custom_setter(client, resource_group_name, server_name, firewall_rule_name, parameters):
    validate_resource_group(resource_group_name)

    endIpAddress = parameters.get('properties', {}).get('endIpAddress')
    startIpAddress = parameters.get('properties', {}).get('startIpAddress')
    parameters = {"properties": {"endIpAddress": endIpAddress, "startIpAddress": startIpAddress}}

    return client.begin_create_or_update(
        resource_group_name,
        server_name,
        firewall_rule_name,
        parameters)


def flexible_firewall_rule_update_custom_func(instance, start_ip_address=None, end_ip_address=None):
    if start_ip_address is not None:
        instance.start_ip_address = start_ip_address
    if end_ip_address is not None:
        instance.end_ip_address = end_ip_address
    return instance


def firewall_rule_get_func(cmd, client, resource_group_name, server_name, firewall_rule_name):
    validate_resource_group(resource_group_name)
    validate_public_access_server(cmd, resource_group_name, server_name)
    return client.get(resource_group_name, server_name, firewall_rule_name)


def firewall_rule_list_func(cmd, client, resource_group_name, server_name):
    validate_resource_group(resource_group_name)
    validate_public_access_server(cmd, resource_group_name, server_name)
    return client.list_by_server(resource_group_name, server_name)


def create_firewall_rule(db_context, cmd, resource_group_name, server_name, start_ip, end_ip):
    validate_resource_group(resource_group_name)

    # allow access to azure ip addresses
    cf_firewall = db_context.cf_firewall  # NOQA pylint: disable=unused-variable
    firewall_client = cf_firewall(cmd.cli_ctx, None)
    firewall = firewall_rule_create_func(cmd=cmd,
                                         client=firewall_client,
                                         resource_group_name=resource_group_name,
                                         server_name=server_name,
                                         start_ip_address=start_ip, end_ip_address=end_ip)
    return firewall.result().name
